import json
import logging
import os
import uuid

from malevich_coretools import (
    AUTH,
    AppLogsWithResults,
    Cfg,
    Endpoint,
    EndpointOverride,
    FlattenAppLogsWithResults,
    UserConfig,
    create_collection_from_df,
    create_doc,
    get_cfg,
    get_endpoint,
    post_collection_object_presigned_url,
    run_endpoint,
    update_collection_object,
)
from malevich_coretools.secondary.const import ENDPOINTS_RUN

from malevich._utility import upload_zip_asset

from .overrides import AssetOverride, CollectionOverride, DocumentOverride, Override

endpoint_logger = logging.getLogger('malevich.run_endpoint')

class MetaEndpoint(Endpoint):
    conn_url: str
    auth: AUTH

    @staticmethod
    def connect_by_hash(hash: str, *, conn_url: str, auth: AUTH) -> "MetaEndpoint":
        endpoint = get_endpoint(
            hash=hash,
            conn_url=conn_url,
            auth=auth
        )
        return MetaEndpoint(
            **endpoint.model_dump(),
            conn_url=conn_url,
            auth=auth
        )

    def get_url(self) -> str:
        return f'{self.conn_url.rstrip("/")}/{ENDPOINTS_RUN(self.hash)}'

    def run(
        self,
        overrides: dict[str, Override] = {},
        endpoint_override: EndpointOverride | None = None,
        with_auth: bool = False,
        *args,
        **kwargs
    ) -> AppLogsWithResults | FlattenAppLogsWithResults:
        """Runs the endpoint with the given override.

        Overrides are tables

        Args:
            endpoint_override (EndpointOverride): Override for the endpoint.
            with_auth (bool, optional): Whether to authenticate the author of the run.
                Defaults to False.

        Returns:
            AppLogsWithResults | FlattenAppLogsWithResults
        """
        assert endpoint_override or overrides, (
            'Either `overrides` or `endpoint_override` should be provided'
        )
        config = Cfg(**json.loads(get_cfg(
            self.taskId,
            conn_url=self.conn_url,
            auth=self.auth,
        )))

        override_keys = list(config.collections.keys())
        diff = set(overrides.keys()) - set(override_keys)

        if diff:
            endpoint_logger.warning(
                'Overrides %s are not in the endpoint config', diff
            )

        overrides = {
            k: v
            for k, v in overrides.items()
            if k in override_keys
        }

        real_overrides = {}
        for key, override in overrides.items():
            if isinstance(override, CollectionOverride):
                real_overrides[key] = create_collection_from_df(
                    override.data,
                    conn_url=self.conn_url,
                    auth=self.auth,
                )

                endpoint_logger.info(
                    'Override collection %s: %s', key, real_overrides[key]
                )

            elif isinstance(override, DocumentOverride):
                real_overrides[key] = '#' + create_doc(
                    override.data.model_dump_json(),
                    conn_url=self.conn_url,
                    auth=self.auth,
                )

                endpoint_logger.info(
                    'Override document %s: %s', key, real_overrides[key]
                )
            elif isinstance(override, AssetOverride):
                if override.file:
                    if not override.path:
                        override.path = (
                            'overrides/' +
                            uuid.uuid4().hex +
                            '/' +
                            os.path.basename(override.file)
                        )

                    update_collection_object(
                        override.path,
                        override.file,
                        conn_url=self.conn_url,
                        auth=self.auth,
                    )
                else:
                    if not override.path:
                        override.path = 'overrides/' + uuid.uuid4().hex

                    url = post_collection_object_presigned_url(
                        override.path,
                        expires_in=600,
                        conn_url=self.conn_url,
                        auth=self.auth,
                    )

                    upload_zip_asset(
                        url,
                        files=override.files,
                        folder=override.folder,
                    )
                real_overrides[key] = '$' + override.path

                endpoint_logger.info(
                    'Override asset %s: %s', key, real_overrides[key]
                )


        if endpoint_override is None:
            endpoint_override = EndpointOverride(
                cfg=UserConfig(collections=real_overrides),
                withResult=True,
                formatLogs=True,
            )

        return run_endpoint(
            self.hash,
            endpoint_override,
            conn_url=self.conn_url,
            with_auth=with_auth,
            *args,
            **kwargs
        )





import logging

from gql import gql
from malevich_space.ops import SpaceOps
from malevich_space.schema import SpaceSetup

from malevich._db import cache_user, get_cached_users, get_db
from malevich._db.schema.core_creds import CoreCredentials
from malevich.core_api import check_auth

logger = logging.getLogger('malevich.core.auto_creds')

def get_core_creds_from_setup(setup: SpaceSetup) -> tuple[str, str]:
    ops = SpaceOps(setup)
    org = setup.org

    if org is not None:
        org = ops.get_org(reverse_id=org).uid

    if creds_ := get_cached_users(
        email=setup.username,
        api_url=setup.api_url,
        host=setup.host.conn_url,
        org_id=org
    ):
        try:
            check_auth(auth=creds_, conn_url=setup.host.conn_url)
            logger.debug(f"Cache hit: {creds_[0]}")
        except Exception:
            pass
        else:
            return creds_


    r = ops.client.execute(
        gql("""
        query GetAllHosts {
            hosts {
                public {
                    edges {
                        node {
                            details {
                                uid
                                connUrl
                            }
                        }
                    }
                }
            }
        }
        """)
    )

    hosts = [
        (
            x['node']['details']['connUrl'],
            x['node']['details']['uid']
        )
        for x in r['hosts']['public']['edges']
    ]

    uid = None
    for conn_url, uid_ in hosts:
        if conn_url == setup.host.conn_url:
            uid = uid_
            break
    else:
        raise Exception("Host not found")

    r = ops.client.execute(
        gql("""
            query GetAllSAs($host_id: String!) {
                host(uid: $host_id) {
                    mySaOnHost (first: 10000000){
                        edges {
                            node {
                                details {
                                    corePassword
                                    coreUsername
                                }
                            }
                        }
                    }
                }
            }
        """), variable_values={'host_id': uid}
    )
    sas = [
        (
            x['node']['details']['coreUsername'],
            x['node']['details']['corePassword']
        )
        for x in r['host']['mySaOnHost']['edges']
    ]
    for u, p in sas:
        parts_ = u.split('__')
        if (
            len(parts_) == 2 and parts_[0] == org
            or len(parts_) == 1 and org is None
        ):
            try:
                check_auth(auth=(u, p,), conn_url=setup.host.conn_url)
            except Exception:
                logger.debug(f"panic: invalid SA {u}, {setup.model_dump()}")
                raise Exception("SA found but not authorizing")
            cache_user(
                email=setup.username,
                api_url=setup.api_url,
                host=setup.host.conn_url,
                org_id=org,
                core_username=u,
                core_password=p
            )
            return u, p
    else:
        raise Exception("SA not found")

def get_core_creds_from_db(user: str, host: str):
    creds_ = get_db().query(CoreCredentials).where(
        CoreCredentials.user == user,
        CoreCredentials.host == host
    ).one_or_none()

    if not creds_:
        return None

    return creds_.user, creds_.password



from typing import Iterator, Optional
from uuid import uuid4

from pydantic import BaseModel, Field

from malevich.models.python_string import PythonString


class BaseNode(BaseModel):
    uuid: str = Field(default_factory=lambda: uuid4().hex)
    alias: Optional[PythonString] = Field(default_factory=lambda: None)

    def __eq__(self, other: "BaseNode") -> bool:
        if not isinstance(other, BaseNode):
            return False
        return self.uuid == other.uuid

    def get_senstivite_fields(self) -> dict[str, str]:
        return {}

    def set_sensitive_fields(self, values: dict[str, str]) -> None:
        pass

    def short_info(self) -> str:
        return f"{self.__class__.__name__}({self.uuid[:6]}, {self.alias})"

    def __hash__(self) -> int:
        return hash(self.uuid)

    def __iter__(self) -> Iterator[tuple[dict[str, bool] | None, 'BaseNode']]:
        return iter({None: self}.items())



"""
AutoFlow

The package provides a context manager for providing
an access to an execution tree in function contexts.
"""

from typing import Iterable

from malevich._dev.singleton import SingletonMeta

from .tree import ExecutionTree


class Flow(metaclass=SingletonMeta):
    """Context-manager that holds the execution tree

    The class is a singleton and can be used as a context-manager
    to create a new execution tree. It holds a stack of trees to
    maintain nested flows. The stack is thread-safe.

    Example:

    .. code-block: python

        from malevich._autoflow.flow import Flow

        with Flow() as tree:
            # Work with tree
            print("Called in flow: ",  Flow.isinflow())
            print("Trees: ", len(Flow.get_stack()))
            print("In flow", Flow.get_stack(), Flow.flow_ref())
            with Flow() as inner:
                print("Trees (with inner): ", len(Flow.get_stack()))
                print("In inner", Flow.get_stack(), Flow.flow_ref())
            print("After exit", Flow.get_stack(), Flow.flow_ref())

        print("At the end", Flow.get_stack())
    """

    # Internal representation of the execution tree
    __flow_stack = []

    def isinflow() -> bool:
        """Check if the current execution is inside a flow"""
        return len(Flow.__flow_stack) > 0

    def flow_ref() -> ExecutionTree:
        """Returns the reference to the current flow in the context"""
        return None if not Flow.isinflow() else Flow.__flow_stack[-1]

    def __enter__(self) -> ExecutionTree:
        Flow.__flow_stack.append(ExecutionTree())
        return Flow.flow_ref()

    def __exit__(self, *args: Iterable) -> None:
        Flow.__flow_stack.pop()

    @staticmethod
    def get_stack() -> tuple[ExecutionTree]:
        """Returns immutable stack of execution trees"""
        return tuple(Flow.__flow_stack)

    @staticmethod
    def _hardset_tree(tree: ExecutionTree) -> None:
        """Injects a tree into the flow stack"""
        Flow.__flow_stack[-1] = tree




from malevich._utility import FlowStub
from malevich.install.installer import Installer
from malevich.manifest import ManifestManager
from malevich.models.dependency import Integration

manf = ManifestManager()

class FlowInstaller(Installer):

    def install(self, reverse_id: str, integrations: list[Integration]) -> None:
        FlowStub.sync_flows(
            reverse_id,
            integrations=integrations
        )

    def restore(self, reverse_id: str, integrations: list[Integration]) -> None:
        return self.install(
            reverse_id,
            integrations
        )

    def construct_dependency(self, object: dict) -> None:
        return None



import re
from collections import defaultdict
from typing import Literal, Optional, overload
from uuid import uuid4

from malevich_space.ops.component_manager import ComponentManager
from malevich_space.ops.space import SpaceOps
from malevich_space.schema import SpaceSetup, VersionMode
from malevich_space.schema.asset import CreateAsset
from malevich_space.schema.branch import BranchSchema
from malevich_space.schema.cfg import CfgSchema
from malevich_space.schema.collection_alias import CollectionAliasSchema
from malevich_space.schema.component import (
    ComponentSchema,
    LoadedComponentSchema,
)
from malevich_space.schema.flow import (
    FlowSchema,
    InFlowAppSchema,
    InFlowComponentSchema,
    InFlowDependency,
    LoadedFlowSchema,
    OpSchema,
    Terminal,
)
from malevich_space.schema.schema import SchemaMetadata

from malevich._autoflow import traced, tracedLike
from malevich._utility import (
    CacheManager,
    LogLevel,
    Registry,
    cout,
    get_space_leaves,
    resolve_setup,
    unique,
    upload_zip_asset,
)
from malevich.interpreter import Interpreter
from malevich.models import (
    Action,
    ArgumentLink,
    BaseNode,
    BaseTask,
    CollectionNode,
    InterpretationError,
    OperationNode,
    SpaceInterpreterState,
    SpaceTask,
    TreeNode,
    VerbosityLevel,
)

from .._autoflow import tracer as gn
from ..manifest import ManifestManager
from ..models.nodes.asset import AssetNode
from ..types import TracedNode

manf = ManifestManager()
reg = Registry()
cache = CacheManager()

_levels = [LogLevel.Info, LogLevel.Warning, LogLevel.Error, LogLevel.Debug]
_actions = [Action.Interpretation, Action.Preparation, Action.Run, Action.Results]


def _log(
    message: str,
    level: int = 0,
    action: int = 0,
    step: bool = False,
    *args,
) -> None:
    cout(
        _actions[action],
        message,
        verbosity=VerbosityLevel.AllSteps if step else VerbosityLevel.OnlyStatus,
        level=_levels[level],
        *args,
    )


names_ = defaultdict(lambda: 0)


def _name(base: str) -> int:
    names_[base] += 1
    return names_[base]


class SpaceInterpreter(Interpreter[SpaceInterpreterState, SpaceTask]):
    """
    Interpret flows to be added and uploaded to your Malevich Space workspace.

    .. note::

        The interpreter can only operate with dependencies installed with
        both `space` installer.

    .. warning::

        Interpreting flows with this interpreter will immediately upload
        the flow to your Malevich Space workspace. Use :meth:`.interpret`
        wisely.

    Interpretation is equivalent to the Add
    action or :code:`malevich space component add` command.

    Prepare
    ---------

    Preparation creates a deployment of the flow in the Malevich Space.
    It is equialent to the Build action or :code:`malevich space component build` and
    :code:`malevich space component boot` commands.

    Run
    -----

    Run executes the flow in the Malevich Space. It is equivalent to the
    Run action or :code:`malevich space component run` command.

    Stop
    ------

    Deletes a deployment of the flow in the Malevich Space. It is equivalent
    to the Stop action or :code:`malevich space component stop` command.

    Results
    -------

    Results is represented as a list of
    :class:`malevich.results`
    objects.
    """

    supports_subtrees = True

    def prettify_collection_id(self, collection_id: str) -> str:
        # Replace spaces with dashes
        _s = re.sub(r"\s+", "-", collection_id)
        # Lowercase
        _s = _s.lower()
        # Replace underscores with dashes
        return _s

    def prettify_collection_name(self, collection_name: str) -> str:
        # Replace multiple spaces with one
        _s = re.sub(r"\s+", " ", collection_name)
        # Replace dashes and underscores with spaces
        _s = re.sub(r"-", " ", _s)
        # Title case
        return _s.title()

    def prettify_schema_id(self, schema_name: str) -> str:
        # 'Schema Name - Some_Name' -> 'SchemaNameSome_Name"
        _s = re.sub(r"[\s-]+", " ", schema_name)
        return _s.replace(" ", "").lower()

    def prettify_component_name(self, component_name: str) -> str:
        _s = re.sub(r"[\s-]+", " ", component_name)
        return _s.title()

    def prettify_config_name(self, component_name: str) -> str:
        _s = re.sub(r"[\s-]+", " ", component_name)
        return "Meta Config for " + _s.title()

    def prettify_config_id(
        self,
        component_name: str,
        rand: Optional[str] = None,  # Random string
    ) -> str:
        rand = rand or uuid4().hex[:8]
        __b = "meta-config-for-" + re.sub(r"[\s|_|-]+", " ", component_name)
        if rand:
            __b += "-" + rand
        return __b

    def update_state(self, state: SpaceInterpreterState = None) -> None:
        """
        The state contains pydantic models, which are not deepcopyable.
        So, this method is used to update the state using custom `copy` method.
        """
        if state:
            self._state = state.copy()
        return state

    @property
    def state(self) -> SpaceInterpreterState:
        return self._state

    def __init__(
        self,
        setup: SpaceSetup | None = None,
        ops: SpaceOps | None = None,
        version_mode: VersionMode = VersionMode.PATCH,
        update_collections: bool = False,
    ) -> None:
        """Space Interpreter

        Args:
            name (str): Name of the flow.
            reverse_id (str): Reverse id of the flow.
        """
        super().__init__()
        self._state = SpaceInterpreterState()
        if not setup and not ops:
            try:
                setup = resolve_setup(manf.query("space", resolve_secrets=True))
            except Exception as e:
                from malevich._cli import login
                if login():
                    setup = resolve_setup(manf.query("space", resolve_secrets=True))
                raise InterpretationError(
                    "Failed to resolve space setup. "
                    "Please check your manifest file.",
                    self,
                    self._state,
                ) from e

        space = ops or SpaceOps(setup)

        try:  # Local patch for space
            host_ = space.get_my_hosts(url=setup.host.conn_url)[0]
        except Exception:
            host_ = None

        if not host_:
            try:
                host_ = space.create_host(
                    alias=setup.host.alias or "", conn_url=setup.host.conn_url
                )
            except KeyError as ke:
                if "details" in ke.args[0]:
                    # Local patch for space
                    pass

        self._state.component_manager = ComponentManager(
            host=host_, space=space, comp_dir="./"
        )

        self._state.host = host_
        self._state.space = space
        self._upload_collections = update_collections
        self._version_mode = version_mode

        self.update_state()

    @overload
    def _upload_schema(
        self,
        state: SpaceInterpreterState,
        node: CollectionNode,
        skip_create: Literal[True],
    ) -> tuple[str | None, str]:
        pass

    @overload
    def _upload_schema(
        self,
        state: SpaceInterpreterState,
        node: CollectionNode,
        skip_create: Literal[False],
    ) -> tuple[str, str]:
        pass

    def _upload_schema(
        self,
        state: SpaceInterpreterState,
        node: CollectionNode,
        skip_create: bool = False,
    ) -> tuple[str | None, str]:
        if not node.scheme:
            raise InterpretationError(
                "To be able to use collection, it should have an attached schema",
                self,
                state,
            )

        # Understanding scheme
        if isinstance(node.scheme, SchemaMetadata):
            # If scheme is already parsed
            core_id = node.scheme.core_id

            if not (s := state.space.get_schema(core_id=core_id)):
                # HACK: This will fall into another
                # if branch where the scheme will be
                # uploaded
                node.scheme = node.scheme.schema_data
            else:
                schema_uid = s.uid
                return schema_uid, core_id

        if isinstance(node.scheme, str):
            # If scheme is a string, then it should firstly
            # be uploaded
            core_id = self.prettify_schema_id(node.collection.collection_id)

            if skip_create:
                return None, core_id

            schema_uid = state.space.create_scheme(
                # $core_id: String!, $raw: String!, $name: String
                core_id=core_id,
                raw=node.scheme,
                name=self.prettify_schema_id(node.collection.collection_id),
            )
            # RFC: Should I allow to upload collection without scheme?
            assert schema_uid, "Failed to upload the schema"

        return schema_uid, core_id

    def _upload_collection(
        self,
        state: SpaceInterpreterState,
        node: CollectionNode,
        core_id: str | None = None,
    ) -> str:
        """Uploads collection (and underlying scheme) to the space."""
        if not core_id:
            core_id = node.collection.collection_id

        _, schema_core_id = self._upload_schema(state, node)
        # Uploading collection
        alias = self.prettify_collection_name(node.collection.collection_id)

        # RFC!: Is it correct way to upload collection?
        uid = state.space.create_collection(
            host_id=state.host.uid,
            core_id=core_id,
            core_alias=alias,
            schema_core_id=schema_core_id,
            docs=[
                row.to_json() for _, row in node.collection.collection_data.iterrows()
            ],
        )

        return uid

    def _ensure_aliases(self, node: BaseNode):
        if isinstance(node, TreeNode):
            if not node.alias:
                node.alias = unique(node.reverse_id)
            for from_, to_, _ in node.tree.tree:
                self._ensure_aliases(from_.owner)
                self._ensure_aliases(to_.owner)
        else:
            if node.alias is None:
                if isinstance(node, CollectionNode):
                    node.alias = unique(node.collection.collection_id)
                elif isinstance(node, OperationNode):
                    node.alias = unique(node.processor_id)
                elif isinstance(node, AssetNode):
                    node.alias = unique(node.name)
                else:
                    raise InterpretationError(
                        "Unexpected node type: " + str(type(node)),
                        self,
                        self.state,
                    )


    def interpret(self, node: TreeNode, component: ComponentSchema) -> BaseTask:
        self._state.aux.tree = node
        if self._version_mode == VersionMode.DEFAULT:
            loaded = self.state.space.get_parsed_component_by_reverse_id(
                reverse_id=component.reverse_id
            )

            task = SpaceTask(state=self.state, component=loaded)
            all = {i.uid for i in loaded.flow.components}
            non_leaves = set().union(
                *[{i.uid for i in c.prev} for c in loaded.flow.components]
            )
            leaves = all.difference(non_leaves)
            returned = [
                gn.tracedLike(BaseNode(uuid=c.uid, alias=c.alias))
                for c in loaded.flow.components
                if c.uid in leaves
            ]
            task.commit_returned(returned)
            return task

        self._ensure_aliases(node)
        return super().interpret(node, component)

    def before_interpret(self, state) -> SpaceInterpreterState:
        state.flow = FlowSchema()
        return state

    def create_node(
        self, state: SpaceInterpreterState, node: gn.traced[BaseNode]
    ) -> SpaceInterpreterState:
        """Wraps node in the execution graph with a Malevich Space component schema.

        Args:
            state (SpaceInterpreterState): Interpreter state.
            node (gn.traced[BaseNode]): Node to wrap.

        Returns:
            SpaceInterpreterState: Interpreter state.
        """

        if isinstance(node.owner, CollectionNode):  # If the node is a collection
            # Try to get the collection component from the space
            component = state.space.get_parsed_component_by_reverse_id(
                # Using the collection id as the reverse id
                reverse_id=node.owner.collection.collection_id
            )

            if component is None:
                # Wrap the collection in a component schema
                name = self.prettify_collection_name(
                    node.owner.collection.collection_id
                )

                _, schema_core_id = self._upload_schema(
                    state=state, node=node.owner, skip_create=False
                )

                if node.owner.collection.collection_data is None:
                    # NOTE: path is None is the only
                    # case for the exception. However,
                    # there may be some more, which should be
                    # checked as well
                    raise InterpretationError(
                        "Cannot interpret collection "
                        + node.owner.collection.collection_id
                        + ". There is no such collection on Space and"
                        + " there is no data provided to upload. Either use"
                        + " existing collections, or provide `df` or `file` argument"
                        + " in collection(...) call"
                    )

                comp = ComponentSchema(
                    name=name,
                    description=f"Meta collection {name}",
                    reverse_id=node.owner.collection.collection_id,
                    collection=CollectionAliasSchema(
                        core_alias=node.owner.collection.collection_id,
                        schema_core_id=schema_core_id,
                        # NOTE: 0.4.18 patch: path -> docs
                        # path=path,
                        docs=[
                            row.to_json() for _, row in
                            node.owner.collection.collection_data.iterrows()
                        ],
                    ),
                )
            else:
                if self._upload_collections:
                    coll_id = self.prettify_collection_id(
                        node.owner.collection.collection_id
                    )
                    uid = self._upload_collection(
                        state,
                        node.owner,
                        core_id=f"override-{coll_id}-{state.interpretation_id}",
                    )
                    state.collection_overrides[component.collection.uid] = uid

                comp = ComponentSchema(
                    reverse_id=component.reverse_id,
                    name=component.name,
                )

        elif isinstance(node.owner, OperationNode):  # If the node is an operation

            # All the required information to interpret the node
            # should be in the registry
            extra = reg.get(node.owner.operation_id)

            # If no extra information is found, it means
            # local information is too old
            # (but it is user fault, not mine)
            if "reverse_id" not in extra:
                raise InterpretationError(
                    "Could not find reverse_id for the operation. "
                    "Most probably, you are trying to use an operation "
                    "installed by another installer.",
                    self,
                    state,
                )

            # Get the component from the space (it should be there)
            component = (
                state.component_manager.space.get_parsed_component_by_reverse_id(
                    reverse_id=extra["reverse_id"]
                )
            )

            if not component:
                # How then you install it? Was it wiped? :(
                raise InterpretationError(
                    "Trying to interpret an operation that is not installed "
                    "properly or extremely outdated. Try to reinstall "
                    "the operation with\n\t"
                    f"`malevich install {extra['reverse_id']} --using space`",
                    self,
                    state,
                )

            state.components_config[node.owner.uuid] = node.owner.config
            comp = ComponentSchema(
                name=component.name, reverse_id=component.reverse_id, app=component.app
            )
            state.node_to_operation[node.owner.uuid] = node.owner.operation_id

        elif isinstance(node.owner, TreeNode):
            if not node.owner.integrated:
                if not state.children_states.get(node.owner.uuid, None):
                    child_interpreter = SpaceInterpreter(
                        setup=self.state.space.space_setup,
                    )

                    comp = ComponentSchema(
                        name=node.owner.name,
                        reverse_id=node.owner.reverse_id,
                    )

                    task_ = child_interpreter.interpret(node.owner, comp)
                    task_.upload()
                    child_state: SpaceInterpreterState = child_interpreter.state
                    comp.flow = child_state.flow

                    state.children_states[node.owner.uuid] = child_state
                else:
                    child_state = state.children_states[node.owner.uuid]
                    comp = ComponentSchema(
                        name=node.owner.name,
                        reverse_id=node.owner.reverse_id,
                        flow=child_state.flow,
                    )
            else:
                comp = state.space.get_parsed_component_by_reverse_id(
                    reverse_id=node.owner.reverse_id
                )

        elif isinstance(node.owner, AssetNode):
            try:
                comp = self.state.space.get_parsed_component_by_reverse_id(
                    reverse_id=node.owner.name
                )
            except Exception:
                comp = None

            if comp is None or comp.asset.checksum != node.owner.magic():
                asset_ = self.state.space.create_asset(
                    asset=CreateAsset(
                        core_path=node.owner.core_path,
                        is_composite=node.owner.is_composite,
                        checksum=node.owner.magic()
                    )
                )
                comp = ComponentSchema(
                    name=node.owner.name,
                    reverse_id=node.owner.name,
                    asset=asset_
                )

                asset_comp = self.state.space.parse_raw(comp, VersionMode.PATCH)
                asset_comp = self.state.space.get_parsed_component_by_reverse_id(
                    reverse_id=asset_comp.reverse_id
                )

                upload_zip_asset(
                    url=asset_comp.asset.upload_url,
                    file=node.owner.real_path if isinstance(node.owner.real_path, str) else None,
                    files=node.owner.real_path if isinstance(node.owner.real_path, list) else None,
                )




        if not comp:
            raise InterpretationError(
                "Failed to interpret the node. This is a bug, please report it.",
            )
        state.components_alias[node.owner.uuid] = node.owner.alias
        state.components[node.owner.uuid] = comp
        return state

    def create_dependency(
        self,
        state: SpaceInterpreterState,
        from_node: TracedNode,
        to_node: traced[OperationNode],
        link: ArgumentLink[traced[BaseNode]],
    ) -> SpaceInterpreterState:
        """Creates a dependency between two nodes."""

        if isinstance(to_node.owner, TreeNode) and not isinstance(
            from_node.owner, TreeNode
        ):
            inter_flow_map = {}

            if to_node.owner.integrated:
                bridges = link.compressed_nodes

                for _, to in bridges:
                    inter_flow_map[from_node.owner.alias] = to.owner.alias

                dependency = InFlowDependency(
                    from_op_id=(
                        # provided by space installer
                        from_node.owner.operation_id
                        if isinstance(from_node.owner, OperationNode)
                        else None
                    ),
                    to_op_id=to_node.owner.underlying_node.operation_id
                    if isinstance(to_node.owner.underlying_node, OperationNode)
                    else None,
                    alias=from_node.owner.alias,
                    order=link.index,
                    terminals=[
                        Terminal(src=x, target=y) for x, y in inter_flow_map.items()
                    ],
                )
            else:
                caller_alias = from_node.owner.alias

                bridges = link.compressed_nodes

                for _, to in bridges:
                    inter_flow_map[caller_alias] = link.shadow_collection.owner.alias


                dependency = InFlowDependency(
                    from_op_id=(
                        # provided by space installer
                        from_node.owner.operation_id
                        if isinstance(from_node.owner, OperationNode)
                        else None
                    ),
                    to_op_id=to_node.owner.underlying_node.operation_id
                    if isinstance(to_node.owner.underlying_node, OperationNode)
                    else None,
                    alias=from_node.owner.alias,
                    order=link.index,
                    terminals=[
                        Terminal(src=x, target=y) for x, y in inter_flow_map.items()
                    ],
                )

        elif isinstance(from_node.owner, TreeNode) and not isinstance(
            to_node.owner, TreeNode
        ):
            if from_node.owner.integrated:
                dependency = InFlowDependency(
                    from_op_id=from_node.owner.underlying_node.operation_id
                    if isinstance(from_node.owner.underlying_node, OperationNode)
                    else None,
                    to_op_id=to_node.owner.operation_id,
                    alias=from_node.owner.alias,
                    order=link.index,
                    terminals=[
                        Terminal(
                            src=from_node.owner.alias,
                            target=to_node.owner.alias,
                        )
                    ]
                )
            else:
                callee_alias = to_node.owner.alias
                op: OperationNode = from_node.owner.underlying_node
                inter_flow_map = {op.alias: callee_alias}

                dependency = InFlowDependency(
                    from_op_id=op.operation_id,
                    to_op_id=(
                        # provided by space installer
                        to_node.owner.operation_id
                        if isinstance(to_node.owner, OperationNode)
                        else None
                    ),
                    alias=from_node.owner.alias,
                    order=link.index,
                    terminals=[
                        Terminal(src=x, target=y) for x, y in inter_flow_map.items()
                    ],
                )

        elif isinstance(from_node.owner, TreeNode) and isinstance(to_node.owner, TreeNode):  # noqa: E501
            left_op = from_node.owner.underlying_node
            right_edges = link.compressed_nodes
            for rel, right_node in right_edges:
                state.dependencies[to_node.owner.uuid].append(
                    InFlowDependency(
                        from_op_id=(
                            left_op.operation_id
                            if isinstance(left_op, OperationNode)
                            else None
                        ),
                        to_op_id=(
                            right_node.owner.operation_id
                            if isinstance(right_node.owner, OperationNode)
                            else None
                        ),
                        alias=from_node.owner.alias,
                        order=rel.index,
                        terminals=[
                            Terminal(
                                src=left_op.alias,
                                target=right_node.alias,
                            )
                        ],
                    )
                )
            return state
        else:
            dependency = InFlowDependency(
                from_op_id=(
                    # provided by space installer
                    from_node.owner.operation_id
                    if isinstance(from_node.owner, OperationNode)
                    else None
                ),
                to_op_id=to_node.owner.operation_id,
                alias=from_node.owner.alias,
                order=link.index,
            )

        state.dependencies[to_node.owner.uuid].append(dependency)

        return state

    def after_interpret(self, state: SpaceInterpreterState) -> SpaceInterpreterState:
        """Finishes the interpretation by adding components to the flow."""
        for uid, component in state.components.items():
            try:
                # I don't know why, but sometimes
                # it fails, so try/exc here
                loaded_component = state.component_manager.component(
                    component, VersionMode.DEFAULT
                )
            except Exception as e:
                # If it fails, try to get the component
                # by reverse id
                loaded_component = (
                    state.component_manager.space.get_component_by_reverse_id(
                        component.reverse_id
                    )
                )

                # No components?(
                # Nothing to do here
                if loaded_component is None:
                    raise InterpretationError(
                        f"Failed to interpret the flow: component {component.name} "
                        "failed to load. ",
                        self,
                        state,
                    ) from e

            # If the component is an operation
            # and has config
            if uid in state.components_config:
                # RFC: Is it correct way to update config?
                space_config = CfgSchema(
                    readable_name=self.prettify_config_name(component.reverse_id),
                    cfg_json=state.components_config.get(uid),
                    core_name=self.prettify_config_id(component.reverse_id),
                )
            else:
                space_config = None

            if uid in state.node_to_operation:
                extra = reg.get(state.node_to_operation[uid], {})
            else:
                extra = {}

            # Add the component to the flow
            state.flow.components = [
                *state.flow.components,
                InFlowComponentSchema(
                    reverse_id=component.reverse_id,
                    alias=self.state.components_alias[uid],
                    depends={dep.alias: dep for dep in state.dependencies[uid]},
                    app=InFlowAppSchema(
                        active_op=[
                            OpSchema(core_id=extra["processor_name"], type="processor")
                        ]
                    )
                    if component.app is not None
                    else None,
                    active_cfg=space_config,
                ),
            ]

        return state

    def get_task(self, state: SpaceInterpreterState) -> BaseTask[SpaceInterpreterState]:
        if self._component is None:
            raise Exception("Expected _component to be not None")

        self._component.flow = state.flow

        def get_component(branch: str | None = None):
            if branch:
                self._component.branch = BranchSchema(name=branch)

            component = state.component_manager.component(
                self._component,
                self._version_mode,
            )
            if self._version_mode != VersionMode.DEFAULT:
                state.space.auto_layout(flow=component.flow.uid)

            return component

        return SpaceTask(state=self.state, get_component=get_component)

    def attach(
        self,
        reverse_id: str | None = None,
        flow_uid: str | None = None,
        deployment_id: str | None = None,
        search_deployments: bool = True,
    ) -> SpaceTask:
        assert (
            reverse_id or deployment_id or flow_uid
        ), "Either reverse_id or deployment_id or uid should be set"

        component: LoadedComponentSchema | None = (
            self._state.space.get_parsed_component_by_reverse_id(
                reverse_id=reverse_id
            )
        )

        if flow_uid and component.flow is not None:
            component.flow = self._state.space.get_flow(flow_uid)

        if component is not None and component.flow is not None:
            if deployment_id:
                self.state.aux.task_id = deployment_id

            elif search_deployments:
                try:
                    tasks = self._state.space.get_deployments_by_flow(
                        flow_id=component.flow.uid,
                        status=["started"]
                    )
                    if len(tasks) > 0:
                        tasks = sorted(tasks, key=lambda x: x.last_runned_at, reverse=True)
                        self.state.aux.task_id = tasks[0].uid
                except:
                    pass


            flow: LoadedFlowSchema = component.flow

            self._state.flow = flow
            self._state.aux.flow_id = flow.uid

            leaves = get_space_leaves(flow)

            # NOTE: there can be more leaves in future
            leave_aliasses = [
                x.alias for x in flow.components if x in leaves
            ]

            leaf_nodes = [
                tracedLike(BaseNode(alias=x))
                for x in leave_aliasses
            ]

        else:
            raise Exception(
                "Could not attach to the flow. Component failed to be parsed."
            )

        task = SpaceTask(state=self._state, component=component)
        task.commit_returned(leaf_nodes)
        return task



from copy import deepcopy
from typing import Annotated, Any

import malevich_coretools as core
from malevich_coretools import Condition, Processor, Result
from pydantic import BaseModel, ConfigDict, Field, SkipValidation

from ..._core.service.service import CoreService
from ..collection import Collection
from ..nodes.asset import AssetNode
from ..nodes.collection import CollectionNode
from ..nodes.document import DocumentNode
from ..nodes.operation import OperationNode


class BaseCoreState(BaseModel):
    """State of the CoreInterpreterV2"""
    model_config = ConfigDict(
        arbitrary_types_allowed=True
    )

    processors: dict[str, Processor] = {}
    """Mapping of operation aliases to processors"""

    conditions: dict[str, Condition] = {}
    """Mapping of condition aliases to processors"""

    collections: dict[str, Collection] = {}
    """Mapping of collection aliases to collections"""

    operation_nodes: dict[str, OperationNode] = {}
    """Mapping of operation aliases to operation nodes"""

    collection_nodes: dict[str, CollectionNode] = {}
    """Mapping of collection aliases to collection nodes"""

    asset_nodes: dict[str, AssetNode] = {}
    """Mapping of asset aliases to asset nodes"""

    document_nodes: dict[str, DocumentNode] = {}
    """Mapping of asset aliases to asset nodes"""

    results: dict[str, list[Result]] = {}
    """Mapping of processor aliases to results"""


from ._autoflow.tracer import traced
from .models.nodes.asset import AssetNode
from .models.nodes.collection import CollectionNode
from .models.nodes.operation import OperationNode


class OpResult(traced[OperationNode]):
    """A class to represent a result of operation

    This serves as an annotation for the return type of the function.
    The result might be a collection (table), a series of collection or
    an asset. All the cases represented by a single object.
    """

class Collection(traced[CollectionNode]):
    """A class to represent a collection (table)"""

class Asset(traced[AssetNode]):
    """A class to represent an asset"""



import pandas as pd
from pandas.io.json import build_table_schema


def pd_to_json_schema(df: pd.DataFrame, format=False) -> dict:
    df = df.infer_objects()
    table_schema = build_table_schema(df)

    schema = {
        'type': 'object',
        'properties': {},
        'required': [],
    }

    for t in table_schema['fields']:
        if t['name'] == 'index':
            continue

        schema['properties'][t['name']] = {
            'type': t['type']
        }


    return schema

def generate_empty_df_from_schema(schema: dict | None) -> pd.DataFrame:
    if schema:
        columns = list(schema['properties'].keys())
    else:
        columns = []
    return pd.DataFrame(columns=columns)



import rich
from typer import Typer

from malevich.core_api import (
    get_user_limits,
    update_user_limits,
)

from ..misc.make import wrap_command

limits_app = Typer(name='limits')


@limits_app.command(name='my')
@wrap_command(get_user_limits)
def list_endpoints(**kwargs) -> None:
    from malevich.core_api import Config
    rich.print("Limits for user: ", Config.CORE_USERNAME)
    lims = get_user_limits(**kwargs)
    rich.print(lims.model_dump())


@limits_app.command(name='set')
@wrap_command(update_user_limits)
def get(**kwargs) -> None:
    update_user_limits(**kwargs)
    return get_user_limits()



from .installer import Installer
from .image import ImageInstaller
from .space import SpaceInstaller
from .flow import FlowInstaller
from .iso import IsolatedSpaceInstaller
from .stub import create_package_stub



import typer

from malevich.manifest import ManifestManager
from malevich.models.preferences import UserPreferences

prefs = typer.Typer(
    help="Manage user preferences",
)

main_keys = UserPreferences.model_fields.keys()


@prefs.command("set", help="Set a preference")
def set_pref(
    key: str = typer.Argument(..., help="Preference key"),
    value: str = typer.Argument(..., help="Preference value"),
) -> None:
    manf = ManifestManager()
    prefs_ = manf.query("preferences")
    prefs = UserPreferences(**prefs_) if prefs_ else UserPreferences()

    keys = key.split(".")
    if keys[0] not in main_keys:
        typer.echo(f"Unknown preference key: {keys[0]}. Available keys are: {', '.join(main_keys)}")  # noqa: E501
        raise typer.Exit(1)

    if len(keys) == 1:
        prefs.__setattr__(keys[0], value)
    else:
        cursor = prefs
        for key in keys[:-1]:
            try:
                if isinstance(cursor, dict):
                    cursor = cursor.__getitem__(key)
                else:
                    cursor = cursor.__getattribute__(key)
            except AttributeError:
                typer.echo(f"Unknown preference subkey: {key} in {'.'.join(keys)} at position {keys.index(key)}")  # noqa: E501
                raise typer.Exit(1)
        if isinstance(cursor, dict):
            cursor.__setitem__(
                keys[-1], type(cursor.__getitem__(keys[-1]))(value))
        else:
            cursor.__setattr__(
                keys[-1], type(cursor.__getattribute__(keys[-1]))(value))
    manf.put("preferences", value=prefs.model_dump())
    typer.echo(f"Set preference {key} to {value}")


@prefs.command("get", help="Get a preference")
def get_pref(
    key: str = typer.Argument(..., help="Preference key"),
) -> None:
    manf = ManifestManager()
    prefs_ = manf.query("preferences")
    prefs = UserPreferences(**prefs_) if prefs_ else UserPreferences()

    keys = key.split(".")
    if keys[0] not in main_keys:
        typer.echo(
            f"Unknown preference key: {keys[0]}. "
            f"Available keys are: {', '.join(main_keys)}"
        )

    if len(keys) == 1:
        typer.echo(prefs.__getattribute__(keys[0]))

    else:
        cursor = prefs
        for key in keys:
            try:
                if isinstance(cursor, dict):
                    cursor = cursor.__getitem__(key)
                else:
                    cursor = cursor.__getattribute__(key)
            except AttributeError:
                typer.echo(f"Unknown preference subkey: {key} in {'.'.join(keys)} at position {keys.index(key)}")  # noqa: E501
                raise typer.Exit(1)

        typer.echo(cursor)



from copy import deepcopy
from typing import Any

import malevich_coretools as core
from pydantic import BaseModel, ConfigDict, Field, SkipValidation

from ..._core.service.service import CoreService
from .base import BaseCoreState


class CoreParams(BaseModel):
    operation_id: str | None = None
    task_id: str | None = None
    core_host: str | None = None
    core_auth: tuple[str, str] | None = None
    base_config: core.Cfg | None = None
    base_config_id: str | None = None

    def __getitem__(self, key: str) -> Any:  # noqa: ANN401
        return getattr(self, key)

    def __setitem__(self, key: str, value: Any) -> None:  # noqa: ANN401
        setattr(self, key, value)

    def __contains__(self, key: str) -> bool:
        return hasattr(self, str(key))


class CoreInterpreterState(BaseCoreState):
    """State of the CoreInterpreterV2"""
    model_config = ConfigDict(
        arbitrary_types_allowed=True
    )

    service: SkipValidation[CoreService] | None = Field(default=None, exclude=True)
    """CoreService instance"""

    params: CoreParams = CoreParams()
    """Interpreter parameters"""



    def __deepcopy__(self, memo=None) -> "CoreInterpreterState":
        # Deepcopy the state without the service
        return CoreInterpreterState(**deepcopy(self.model_dump()), service=self.service)



import hashlib
import io
import uuid
from typing import Optional

import pandas as pd
from pydantic import BaseModel, ConfigDict, Field

from malevich.models.python_string import PythonString


class Collection(BaseModel):
    collection_id: PythonString
    core_id: Optional[str] = None
    collection_data: Optional[pd.DataFrame] = Field(
        pd.DataFrame([]), repr=False
    )

    model_config = ConfigDict(
        arbitrary_types_allowed=True
    )

    persistent: bool = False

    @staticmethod
    def from_file(file: str, id: None = uuid.uuid4()) -> None:
        return Collection(
            collection_id=id,
            collection_data=pd.read_csv(file)
        )

    def magic(self, with_id=None, with_data=None) -> str:
        if with_id is None:
            with_id = True
        if with_data is None:
            with_data = not self.persistent

        buffer = io.BytesIO()
        self.collection_data.to_pickle(buffer)
        buffer.seek(0)
        data_ = buffer.read()

        a = hashlib.sha256(data_).hexdigest()
        b = hashlib.sha256(self.collection_id.encode()).hexdigest()

        if not with_id:
            b = ""
        if not with_data:
            a = ""

        if with_id and with_data:
            return a[:32] + b[:32]
        elif with_id:
            return b
        elif with_data:
            return a

    def verify(self, hash: str, **kwargs) -> bool:
        return self.magic(**kwargs) == hash



import os

from jinja2 import Environment, FileSystemLoader

from ..models.dependency import Integration
from ..path import Paths


class FlowStub:
    @staticmethod
    def sync_flows(
        reverse_id: str,
        integrations: list[Integration],
    )-> None:
        install_path = Paths.module('flows', reverse_id + '.py')
        environment = Environment(
            loader=FileSystemLoader(Paths.templates('metascript', 'flow_stub'))
        )
        with open(install_path, 'w+') as f:
            f.write(
                environment.get_template('body.jinja2').render(
                   reverse_id=reverse_id,
                   integrations=integrations,
                )
            )

        init_path = Paths.module('flows', '__init__.py')
        with open(init_path, 'a+') as f:
            f.write(
                f'from .{reverse_id} import {reverse_id}\n'
            )

    @staticmethod
    def remove_stub(reverse_id: str) -> None:
        os.remove(Paths.module('flows', reverse_id + '.py'))
        with open(Paths.module('flows', '__init__.py')) as f:
            lines = f.readlines()

        with open(Paths.module('flows', '__init__.py'), 'w') as f:
            for line in lines:
                if f'from .{reverse_id} import {reverse_id}' in line:
                    continue
                f.write(line)



from sqlalchemy import create_engine
from sqlalchemy.orm.session import Session
from malevich._db.schema.base import Base
from malevich.path import Paths
from filelock import FileLock

class Write:
    def __init__(self, timeout: int | None = None):
        self._ref = None
        self._lock = FileLock(Paths.lock('.db'))
        self._tmout = timeout

    def __enter__(self):
        self._lock.acquire(timeout=self._tmout)
        try:
            import sqlite3
            # HACK: Create empty db
            sqlite3.connect(Paths.db()).close()
            engine = create_engine("sqlite:///" + Paths.db())
            Base.metadata.create_all(engine)
            self._ref = Session(bind=engine)
        except (Exception, KeyboardInterrupt) as e:
            self._lock.release()
            raise e
        
        return self._ref
    
    def __exit__(self, exc_type, exc_value, traceback):
        self._ref.close()
        self._lock.release()
        self._ref = None
    


from abc import ABC, abstractmethod
from typing import Any, Generic, Optional, TypeVar

from pydantic import BaseModel

T = TypeVar("T")
Key = TypeVar("Key")
# Data = TypeVar("Output")

class BaseInjectable(BaseModel, Generic[T, Key], ABC):
    @abstractmethod
    def get_inject_key(self) -> Key:
        pass

    @abstractmethod
    def get_inject_data(self) -> T:
        pass

class CoreInjectable(BaseInjectable[str, str]):

    collection_id: str
    alias: str
    uploaded_id: Optional[str] = None
    node: Any = None

    def get_inject_data(self) -> str:
        return self.collection_id

    def get_inject_key(self) -> str:
        return self.alias

    def get_uploaded_id(self) -> Optional[str]:
        return self.uploaded_id


class SpaceInjectable(BaseInjectable[str, str]):
    in_flow_id: str | None = None
    snapshot_flow_id: str | None = None
    alias: str | None = None
    reverse_id: str | None = None


    def get_inject_data(self) -> str:
        return self.snapshot_flow_id

    def get_inject_key(self) -> str:
        return self.alias



import re

import rich
import typer
from rich.prompt import Prompt

from malevich._utility.ci.github import DockerRegistry, GithubCIOps
from malevich.constants import IMAGE_BASE
from malevich.constants import SPACE_API_URL as DEFAULT_SPACE_HOST

github = typer.Typer()


@github.command(help="Initialize the Github CI/CD pipeline")
def init(
    interactive: bool = typer.Option(
        False,
        help="Run the initialization wizard",
        show_default=False,
    ),
    repo_name: str = typer.Option(
        None,
        help="Github repository name in the format <username>/<repo-name>",
        show_default=False,
    ),
    github_token: str = typer.Option(
        None,
        help="Github token with access to the repository",
        show_default=False,
    ),
    space_user: str = typer.Option(
        None,
        help="Malevich Space username",
        show_default=False,
    ),
    space_token: str = typer.Option(
        None,
        help="Malevich Space token (password)",
        show_default=False,
    ),
    space_url: str = typer.Option(
        None,
        help="Malevich Space API URL",
        show_default=False,
    ),
    branch: str = typer.Option(
        None,
        help="Branch to setup CI in",
        show_default=False,
    ),
    registry_url: str = typer.Option(
        'ghcr.io',
        help="Docker Image Registry URL, for example `public.ecr.aws` or 'cr.yandex'",
        show_default=False,
    ),
    registry_id: str = typer.Option(
        'owner',
        help="Docker Registry ID",
        show_default=False,
    ),

    image_user: str = typer.Option(
        'USERNAME',
        help="Username to access the Docker Image Registry",
        show_default=False,
    ),
    image_token: str = typer.Option(
        'token',
        help="Password to access the Docker Image Registry",
        show_default=False,
    ),
    org_id: str = typer.Option(
        'empty',
        help="Malevich space organization ID",
        show_default=False
    )
):
    """Initialize the Github CI/CD pipeline."""
    rich.print("Welcome to the Github CI/CD pipeline initialization wizard!")
    ops = GithubCIOps()

    if interactive:
        # Get the repository name
        repo_name = Prompt.ask(
            "Enter the Github repository name in the format "
            "[pale_green1]<username>/<repo-name>[/pale_green1]"
        )

        if not re.match(r"^.*/.*", repo_name):
            rich.print(
                "[red]Invalid repository name format. Please try again.[/red]"
            )
            return
        # Get Github token
        github_token = Prompt.ask(
            "Enter the Github token with access to the repository "
            "[i bright_black](ghp_...)[/i bright_black]"
        )

        space_user = Prompt.ask(
            "Enter the username to access [b deep_pink3]Malevich Space[/b deep_pink3]"
            " [i bright_black]"
            "(leave empty to use access token instead)[/i bright_black]"
        )

        space_token = Prompt.ask(
            f"Enter the {'password' if space_user else 'token'} to access "
            "[b deep_pink3]Malevich Space[/b deep_pink3]"
        )

        space_url = Prompt.ask(
            "Enter the host of [b deep_pink3]Malevich Space[/b deep_pink3] ",
            default=DEFAULT_SPACE_HOST
        )

        branch = Prompt.ask(
            "Enter the branch to run the pipeline on",
            default="main"
        )

        registry_type = Prompt.ask(
            "Enter the Docker Container Registry type",
            choices=[str(x.value) for x in DockerRegistry],
            default=DockerRegistry.PUBLIC_AWS_ECR,
        )

        if registry_type != DockerRegistry.PRIVATE_AWS_ECR:
            registry_url = Prompt.ask(
                "Enter the Docker Container Registry URL ",
                default=IMAGE_BASE.split('/')[0]
            )

        registry_id = Prompt.ask(
            "Enter the Docker Container Registry ID ",
            default=IMAGE_BASE.split('/')[1]
        )

        image_user = Prompt.ask(
            "Enter the username to access the Docker Container Registry",
        )

        image_token = Prompt.ask(
            "Enter the password to access the Docker Container Registry",
        )

    rich.print("Calling the Github API to setup the CI/CD pipeline")

    ops.setup_ci(
        token=github_token,
        repository=repo_name,
        space_user=space_user,
        space_token=space_token,
        space_url=space_url,
        branch=branch,
        registry_url=registry_url,
        registry_id=registry_id,
        image_user=image_user,
        image_token=image_token,
        org_id=org_id,
        verbose=True
    )

app = typer.Typer()
app.add_typer(github, name="github", help="Github CI/CD pipeline")



from functools import partial, wraps

import malevich_coretools as api

from malevich._utility.core_logging import IgnoreCoreLogs

from ..refs import BaseRef
from .base import BaseCoreService


def update_doc_by_name(name, *args, **kwargs):
    with IgnoreCoreLogs():
        return api.update_doc(
            id=api.get_doc_by_name(
                name=name,
                auth=kwargs["auth"],
                conn_url=kwargs["conn_url"]
            ).id,
            *args,
            **kwargs
        )

def delete_doc_by_name(name, *args, **kwargs):
    return api.delete_doc(
        id=api.get_doc_by_name(
            name=name,
            auth=kwargs["auth"],
            conn_url=kwargs["conn_url"]
        ).id,
        *args,
        **kwargs
    )

class DocumentService(BaseCoreService):
    def __init__(self, auth: api.AUTH, conn_url: str) -> None:
        super().__init__(auth, conn_url)

    def id(
        self,
        id: str,
        /,
    ):
        return BaseRef(
            name=f'DocumentByIdRef({id})',
            create=None,
            delete=partial(
                api.delete_doc,
                id, auth=self.auth, conn_url=self.conn_url
            ),
            update=partial(
                api.update_doc,
                id, auth=self.auth, conn_url=self.conn_url
            ),
            get=partial(
                api.get_doc,
                id, auth=self.auth, conn_url=self.conn_url
            ),
            list=None,
        )


    def name(
        self,
        name: str,
        /,
    ):
        return BaseRef(
            name=f'DocumentByNameRef({name})',
            create=partial(
                api.create_doc,
                name=name, auth=self.auth, conn_url=self.conn_url
            ),
            delete=partial(
                delete_doc_by_name,
                name,
                auth=self.auth,
                conn_url=self.conn_url
            ),
            update=partial(
                update_doc_by_name,
                name,
                auth=self.auth,
                conn_url=self.conn_url
            ),
            get=partial(
                api.get_doc_by_name,
                name, auth=self.auth, conn_url=self.conn_url
            ),
            list=None
        )

    def all(self):
        return BaseRef(
            'DocumentAllRef',
            create=None,
            delete=partial(
                api.delete_docs,
                auth=self.auth, conn_url=self.conn_url
            ),
            update=None,
            get=None,
            list=partial(
                api.get_docs,
                auth=self.auth, conn_url=self.conn_url
            )
        )






from typing import Annotated

import rich
import typer
import typer.core

from malevich._utility import FlowStub, PackageManager
from malevich.manifest import ManifestManager

from .prefs import prefs as prefs


def remove(
    component_names: Annotated[list[str], typer.Argument(...)],
) -> None:
    try:
        manf = ManifestManager()
        removed = []
        for component in manf.query('flows'):
           if component in component_names:
                FlowStub.remove_stub(component)
                manf.remove('flows', component)
                removed.append(component)

        for component in component_names:
            if component in removed:
                continue
            manf.remove(
                'dependencies',
                component,
            )
            PackageManager().remove_stub(component)

        rich.print(
            f"[green]Component(s) [b]{', '.join(component_names)}[/b] removed[/green]"
        )
        rich.print(f"Bye, bye [b]{component}[/b]")
    except Exception as e:
        rich.print(
            f"[red]Failed to remove component [b]{component}[/b][/red]")
        rich.print(e)
        return



from sqlalchemy import create_engine
from sqlalchemy.orm.session import Session
from malevich._db.schema.base import Base
from malevich.path import Paths
from filelock import FileLock

class Read:
    def __init__(self):
        self._ref = None

    def __enter__(self):
        import sqlite3
        # HACK: Create empty db
        sqlite3.connect(Paths.db()).close()
        engine = create_engine("duckdb:///" + Paths.db())
        Base.metadata.create_all(engine)
        self._ref = Session(bind=engine)

        return self._ref
    
    def __exit__(self, exc_type, exc_value, traceback):
        self._ref.close()
        self._ref = None
    


import enum
import os
import tempfile

import requests
from git import Repo
from github import PublicKey

from .base import CIOps


class DockerRegistry(enum.Enum):
    OTHER = "other"
    PUBLIC_AWS_ECR = "ecr"
    PRIVATE_AWS_ECR = "ecr-private"
    YANDEX = "yandex"
    GCR = "ghcr"

class _NotSet:
    pass

class GithubCIOps(CIOps):
    def __create_action(self, repo_dir: str, branch: str) -> str:
        os.makedirs(os.path.join(repo_dir, ".github",
                    "workflows"), exist_ok=True)

        action_file_path = os.path.join(
            repo_dir, ".github", "workflows", f"malevich-ci_{branch}.yml"
        )

        template_file_path = os.path.join(
            os.path.dirname(__file__), 'templates', 'malevich-ci.yml'
        )

        with open(action_file_path, "w") as action, \
                open(template_file_path) as template:

            action.write(template.read().format(
                BRANCH_NAME=branch
            ))

        return action_file_path

    def __create_manual_action(self, repo_dir: str, branch: str) -> str:
        os.makedirs(os.path.join(repo_dir, ".github",
                    "workflows"), exist_ok=True)

        action_file_path = os.path.join(
            repo_dir, ".github", "workflows", f"malevich-ci-manual_{branch}.yml"
        )

        template_file_path = os.path.join(
            os.path.dirname(__file__), 'templates', 'malevich-ci-manual.yml'
        )

        with open(action_file_path, "w") as action, \
                open(template_file_path) as template:

            action.write(template.read().format(
                BRANCH_NAME=branch
            ))

        return action_file_path

    def _log(self, verbose: bool, msg: str) -> None:
        if verbose:
            print(msg)

    def setup_ci(
        self,
        token: str,
        repository: str,
        space_user: str,
        space_token: str,
        space_url: str,
        branch: str,
        registry_url: str = 'ghcr.io',
        registry_id: str = _NotSet,
        image_user: str = 'USERNAME',
        image_token: str = _NotSet,
        org_id: str = 'empty',
        verbose: bool = False,
    ) -> None:
        """Setup CI for a Github repository

        Args:
            token (str): Github token
            repository (str): Github repository in the form of `user/repo`
            space_user (str): Malevich Space username
            space_token (str): Malevich Space token (password)
            space_url (str): Malevich Space API URL
            branch (str): Branch to setup CI in
            registry_url (str):
                Docker Image Registry URL, for example `public.ecr.aws` or 'cr.yandex'
            registry_id (str): Docker Registry ID
            registry_type (DockerRegistry): Docker Registry type
            image_user (str): Username to access the Docker Image Registry
            image_token (str): TOKEN to access the Docker Image Registry
            org_id (str): ORGANIZATION_ID
            verbose (bool, optional): Verbose mode. Defaults to False.
        """

        pub_key = requests.get(
            f"https://api.github.com/repos/{repository}/actions/secrets/public-key",
            headers={
                "Accept": "application/vnd.github+json",
                "Authorization": f"Bearer {token}",
                "X-GitHub-Api-Version": "2022-11-28"
            }
        )
        if pub_key.status_code >= 400:
            raise Exception(
                "We couldn't get the public key from Github. "
                "Most probably, given credentials are invalid or "
                "do not have enough permissions. "
                "Below is the response from Github:\n"
                f"{pub_key.json()}"
            )
        pub_key = pub_key.json()

        keys = [
            "SPACE_USERNAME",
            "SPACE_PASSWORD",
            "API_URL",
            "IMAGE_USERNAME",
            "IMAGE_PASSWORD",
            "REGISTRY_URL",
            "REGISTRY_ID",
            "REGISTRY_TYPE",
            "SPACE_ORGANIZATION_ID"
        ]

        registry_type = DockerRegistry.GCR
        if 'ecr' in registry_url:
            registry_type = (
                DockerRegistry.PUBLIC_AWS_ECR if 'public' in registry_url
                else DockerRegistry.PRIVATE_AWS_ECR
            )
        elif 'yandex' in registry_url:
            registry_type = DockerRegistry.YANDEX

        values = [
            space_user,
            space_token,
            space_url,
            image_user,
            token if image_token == _NotSet else image_token,
            registry_url,
            repository.split('/')[0].lower() if registry_id == _NotSet else registry_id,
            registry_type,
            org_id
        ]

        for key, value in zip(keys, values):
            payload = PublicKey.encrypt(pub_key['key'], value)
            url = f'https://api.github.com/repos/{repository}/actions/secrets/{key}'

            requests.put(
                url,
                headers={
                    "Accept": "application/vnd.github+json",
                    "Authorization": f"Bearer {token}",
                    "X-GitHub-Api-Version": "2022-11-28"
                },
                json={
                    'encrypted_value': f'{payload}',
                    'key_id': f'{pub_key["key_id"]}'
                }
            )
            self._log(verbose, f"Updated secret {key}")

        with tempfile.TemporaryDirectory() as tmpdir:
            git_repo = Repo.clone_from(
                f'https://{token}@github.com/{repository}', tmpdir
            )
            self._log(verbose, f"Cloned repository {repository} into {tmpdir}")
            git_repo.git.checkout(branch)
            self._log(verbose, f"Checkouted to branch {branch}")
            action_file = self.__create_action(tmpdir, branch)
            manual_action_file = self.__create_manual_action(tmpdir, branch)
            self._log(verbose, f"Created action file {action_file}")
            git_repo.git.add(action_file)
            git_repo.git.add(manual_action_file)
            try:
                git_repo.git.commit(
                    "-m",
                    "add: malevich-ci.yml && malevich-ci-manual.yml"
                )
                self._log(
                    verbose,
                    f"Committed actions file {action_file}, {manual_action_file}"
                )
                git_repo.git.push()
            except Exception as e:
                self._log(
                    verbose,
                    f"{action_file} and {manual_action_file} already exist"
                )

                print(e)



from collections import deque
from typing import Any, List, Tuple


def flatdict(d: dict) -> List[Tuple[Tuple[str, ...], Any]]:
    """
    Flattens a nested dictionary into a list of tuples,
    where each tuple contains the path to a leaf node and its value.

    Args:
        d (dict): The dictionary to flatten.

    Returns:
        list:
            A list of tuples, where each tuple contains
            the path to a leaf node and its value.
    """
    res = []
    q = deque()

    q.extend([((k,), v) for k, v in d.items()])
    while q:
        path, value = q.pop()
        if isinstance(value, dict):
            q.extend([((*path, k), v) for k, v in value.items()])
        elif isinstance(value, list):
            q.extend([((*path, k), v) for k, v in enumerate(value)])
        else:
            res.append(
                (
                    path,
                    value,
                )
            )
    return res




import pickle

# from .interpreted import InterpretedTask
from enum import Enum
from typing import Any, Type

from malevich_space.schema import ComponentSchema

from malevich.types import FlowOutput

from ..endpoint import MetaEndpoint
from ..nodes.tree import TreeNode
from ..results import Result
from .base import BaseInjectable, BaseTask


class PromisedStage(Enum):
    NOT_INTERPRETED = "NOT_INTERPRETED"

# Proxy Pattern
class PromisedTask(BaseTask[PromisedStage]):
    """An intermediate representation of a task that can be interpreted

    Interpretation of a task means attaching it to a particular platform
    and preparing it for execution. This is done by calling :meth:`interpret`
    method.

    Once the task is interpreted, the following methods become available:

    - :meth:`prepare`
    - :meth:`run`
    - :meth:`stop`
    - :meth:`results`
    """
    def __init__(
        self,
        results: FlowOutput,
        tree: TreeNode,
        component: ComponentSchema
    ) -> None:
        self.__results = results
        self.__tree = tree
        self.__task = None
        self.__conf_memory = []
        self._component = component

    @property
    def tree(self) -> TreeNode:
        return self.__tree

    def _attach_task(self, _task: BaseTask) -> None:
        self.__task = _task

    def get_stage(self) -> Any:  # noqa: ANN401
        """Retrieves the current stage of the task.

        If the task has not been interpreted, the stage is `NOT_INTERPRETED`.
        Otherwise, the stage is defined by the underlying interpreted task.
        """
        if self.__task:
            return self.__task.get_stage()
        else:
            return PromisedStage.NOT_INTERPRETED

    def get_stage_class(self) -> Type:
        """Returns the enum class of the stage of the task

        If the task has not been interpreted, the class is `PromisedStage`.
        Otherwise, the class is defined by the underlying interpreted task.
        """
        if self.__task:
            return self.__task.get_stage_class()
        else:
            return PromisedStage

    def interpret(
        self,
        interpreter: 'malevich._interpreter.abstrct.Interpreter' = None
    ) -> None:
        """Interprets the task with a particular interpreter

        Attaching a task to a particular platform and preparing it for execution.
        This is done by calling :meth:`interpret`. Once the task is interpreted,
        it can be prepared, run, stopped and its results can be fetched.

        The task might be re-interpreted with a different interpreter. In this
        case, the previous interpreter is discarded and the new one is used.

        Args:
            interpreter (:class:`malevich.interpreter.Interpreter`, optional):
                Interpreter to use. :class:`malevich.interprete.SpaceInterpreter`
                is used when not specified. Defaults to None.
        """
        from malevich.interpreter import SpaceInterpreter
        __interpreter = interpreter or SpaceInterpreter()
        try:
            task = __interpreter.interpret(self.__tree, self._component)
            if self.__results:
                task.commit_returned(self.__results)
            self.__task = task
            # Apply the configuration that was stored in memory
            for operation, kwargs in self.__conf_memory:
                self.__task.configure(operation, **kwargs)
            del self.__conf_memory
        except Exception as e:
           if not interpreter:
               raise Exception(
                    "Attempt to interpret task with default interpreter failed. "
                    "Try to specify interpreter explicitly"
                ) from e
           else:
                raise e

    async def prepare(self, *args, **kwargs) -> None:
        """Prepares necessary data for the task to be executed (run)

        Accepts any arguments and keyword arguments and passes them to the
        underlying callback created in the interpreter itself. For particular
        arguments and keyword arguments, see the documentation of the interpreter
        used before calling this method.
        """
        if not self.__task:
            raise Exception(
                "Unable to prepare task, that has not been interpreted. "
                "Please, use `.interpret` first to attach task to "
                "a particular platform"
            )

        return await self.__task.prepare(*args, **kwargs)

    async def run(
        self,
        run_id: str | None = None,
        override: dict[str, Any] | None = None,
        *args,
        **kwargs
    ) -> None:
        """Runs the task

        Accepts any arguments and keyword arguments and passes them to the
        underlying callback created in the interpreter itself. For particular
        arguments and keyword arguments, see the documentation of the interpreter
        used before calling this method.
        """
        if override is None:
            override = dict()

        if not self.__task:
            raise Exception(
                "Unable to run task, that has not been interpreted. "
                "Please, use `.interpret` first to attach task to "
                "a particular platform"
            )
        # TODO: try/except with error on this level
        # if task is not prepared
        return await self.__task.run(
            *args,
            run_id=run_id,
            override=override,
            **kwargs
        )

    async def stop(self, *args, **kwargs) -> None:
        """Stops the task

        Accepts any arguments and keyword arguments and passes them to the
        underlying callback created in the interpreter itself. For particular
        arguments and keyword arguments, see the documentation of the interpreter
        used before calling this method.
        """
        if not self.__task:
            raise Exception(
                "Unable to prepare task, that has not been interpreted. "
                "Please, use `.interpret` first to attach task to "
                "a particular platform"
            )
        # TODO: try/except with error on this level
        # if task is not prepared
        return await self.__task.stop(*args, **kwargs)

    async def results(self, *args, **kwargs) -> list[Result]:
        """Retrieve results of the task

        Accepts any arguments and keyword arguments and passes them to the
        underlying callback created in the interpreter itself. For particular
        arguments and keyword arguments, see the documentation of the interpreter
        used before calling this method.
        """
        if not self.__task:
            raise Exception(
                "Unable to get results of the task, that has not been interpreted. "
                "Please, use `.interpret` first to attach task to "
                "a particular platform"
            )

        return await self.__task.results(*args, **kwargs)

    def commit_returned(self, returned: FlowOutput) -> None:
        """Saves objects to determine the results of the task

        Accepts any arguments and keyword arguments and passes them to the
        underlying callback created in the interpreter itself. For particular
        arguments and keyword arguments, see the documentation of the interpreter
        used before calling this method.
        """
        if not self.__task:
            raise Exception(
                "Unable to commit returned results to the task, "
                "that has not been interpreted. "
                "Please, use `.interpret` first to attach task to "
                "a particular platform"
            )

        return self.__task.commit_returned(returned)

    def dump(self) -> bytes:
        """Serialize the task to bytes, which can be saved and used to load it again"""
        if self.__task:
            task_bytes_ = self.__task.dump()
        else:
            task_bytes_ = b""
        tree_bytes_ = pickle.dumps(self.__tree.model_dump())
        results_bytes_ = pickle.dumps(self.__results)
        component_bytes_ = pickle.dumps(self._component)
        return pickle.dumps(
            (task_bytes_, tree_bytes_, results_bytes_, component_bytes_)
        )

    @staticmethod
    def load(object_bytes: bytes) -> 'PromisedTask':
        """Static method. Deserialize bytes into task object"""
        task_bytes_, tree_bytes_, results_bytes_, component_bytes_ = pickle.loads(object_bytes)  # noqa: E501
        task = PromisedTask(
            results=pickle.loads(results_bytes_),
            tree=TreeNode(**pickle.loads(tree_bytes_)),
            component=pickle.loads(component_bytes_)
        )
        if len(task_bytes_) > 0:
            task._attach_task(pickle.loads(task_bytes_))

        return task


    def get_injectables(self) -> list[BaseInjectable]:
        """Returns a list of possible injections into run

        See the documentation of the corresponding
        interpreter used before calling this method.
        """
        if not self.__task:
            raise Exception(
                "Unable to get injectables. "
                "Please, use `.interpret` first to attach task to "
                "a particular platform"
            )
        return self.__task.get_injectables()

    def get_operations(self) -> list[str]:
        """Returns a list of operations

        The term operations is defined by the interpreter used.
        See the documentation of the corresponding
        interpreter used before calling this method.
        """
        if not self.__task:
            raise Exception(
                "Unable to get operations. "
                "Please, use `.interpret` first to attach task to "
                "a particular platform"
            )
        return self.__task.get_operations()

    def configure(self, *operations: str, **kwargs) -> None:
        """Configures the task for a particular operation

        What is configurable and how it is configurable is defined by the
        interpreter used. See the documentation of the corresponding
        interpreter used before calling this method.

        Args:
            operation (str): Operation to configure (one from :meth:`get_operations`)
            **kwargs (Any): Arguments to configure the operation
        """
        if not self.__task:
            for op in operations:
            # Remember the configuration to apply it later
                self.__conf_memory.append((op, kwargs))
            return None
        # Otherwise proxy directly
        return self.__task.configure(*operations, **kwargs)

    def get_interpreted_task(self) -> BaseTask:
        """Retrieves the interpreted task that is produced when calling :meth:`interpret`

        Returns:
            :class:`BaseTask`: The interpreted task
        """  # noqa: E501
        if not self.__task:
            raise Exception(
                "Unable to get interpreted task. "
                "Please, use `.interpret` first to attach task to "
                "a particular platform"
            )
        return self.__task

    def publish(self, *args, **kwargs) -> MetaEndpoint:
        """Creates a HTTP endpoint for the task

        Accepts any arguments and keyword arguments and passes them to the
        underlying callback created in the interpreter itself. For particular
        arguments and keyword arguments, see the documentation of the interpreter
        used before calling this method.

        Returns:
            :class:`malevich.models.endpoint.MetaEndpoint`: An endpoint object
        """
        if not self.__task:
            raise Exception(
                "Cannot publish uninterpreted task. "
                "Use `.interpret()` first."
            )
        return self.__task.publish(*args, **kwargs)



from ..schema import Base, CachedCredentials
from .get_db import get_db

def get_cached_users(
    email: str,
    api_url: str,
    host: str,
    org_id: str | None
) -> tuple[str, str] | None:
    session = get_db()
    creds: CachedCredentials = session.query(CachedCredentials).filter(
        CachedCredentials.email == email,
        CachedCredentials.host == host,
        CachedCredentials.org_id == org_id,
        CachedCredentials.api_url == api_url
    ).one_or_none()
    
    if creds is not None:
        return creds.core_username, creds.core_password
    else:
        return None
    

def cache_user(
    email: str,
    api_url: str,
    host: str,
    org_id: str | None,
    core_username: str,
    core_password: str
) -> None:
    session = get_db()
    creds: CachedCredentials = session.query(CachedCredentials).filter(
        CachedCredentials.email == email,
        CachedCredentials.host == host,
        CachedCredentials.org_id == org_id,
        CachedCredentials.api_url == api_url
    ).one_or_none()
    if not creds:
        session.add(
            CachedCredentials(
                email=email,
                api_url=api_url,
                host=host,
                org_id=org_id,
                core_username=core_username,
                core_password=core_password
            )
        )
    else:
        creds.core_username = core_username
        creds.core_password = core_password
    session.commit()




import hashlib
from pathlib import Path

from malevich_app import LocalRunner, LocalRunStruct
from malevich._utility.stub import Stub
from malevich.models.installers.local import LocalDependency, LocalOptions
from malevich.models.manifest import Dependency
from malevich.path import Paths

from .installer import Installer


class LocalInstaller(Installer):
    """Installs package using local files"""
    name = "local"

    def install(self, package_name: str, app_path: Path) -> LocalDependency:
        if not app_path.exists():
            raise FileNotFoundError(f"The path {app_path} does not exist")
        if not app_path.is_dir():
            raise NotADirectoryError(f"The path {app_path} is not a directory")

        dependencies = []
        if (app_path / "requirements.txt").exists():
            dependencies = open(app_path / "requirements.txt").readlines()

        dependency = LocalDependency(
            package_id=package_name,
            options=LocalOptions(import_path=app_path, dependencies=dependencies),
        )

        runner = LocalRunner(
            LocalRunStruct(import_dirs=[str(app_path.absolute())])
        )

        operation_ids = {
            str(op.id): hashlib.sha256(op.model_dump_json().encode()).hexdigest()
            for op in [*runner.app_info.processors.values(), *runner.app_info.conditions.values()]
        }

        operation_names = {
            str(op.id): op.name
            for op in [*runner.app_info.processors.values(), *runner.app_info.conditions.values()]
        }

        Stub.from_app_info(
            app_info=runner.app_info,
            path=Paths.module(package_name),
            package_name=package_name,
            dependency=dependency,
            operation_ids=operation_ids,
            registry_records={
                processor_id: {
                    "operation_id": operation_id,
                    "processor_id": operation_names[processor_id],
                    "package_path": str(app_path.absolute())
                }
                for processor_id, operation_id in operation_ids.items()
            },
        )

        return LocalDependency(
            package_id=package_name,
            options=LocalOptions(import_path=app_path, dependencies=dependencies),
        )

    def construct_dependency(self, object: dict) -> Dependency:
        return LocalDependency(**object)

    def restore(self, dependency: Dependency) -> Dependency:
        if not isinstance(dependency, LocalDependency):
            raise ValueError(f"Invalid dependency type: {type(dependency)}")

        return self.install(
            package_name=dependency.package_id,
            app_path=Path(dependency.options.import_path),
        )



import malevich_coretools as api  # noqa: I001

from .base import BaseCoreService
from .asset import AssetService
from .collection import CollectionService
from .configuration import ConfigService
from .pipeline import PipelineService
from .document import DocumentService
from .runs import RunService

class CoreService(BaseCoreService):
    def __init__(self, auth: api.AUTH, conn_url: str) -> None:
        super().__init__(auth, conn_url)

        self.collection = CollectionService(auth, conn_url)
        self.cfg = ConfigService(auth, conn_url)
        self.asset = AssetService(auth, conn_url)
        self.pipeline = PipelineService(auth, conn_url)
        self.document = DocumentService(auth, conn_url)
        self.run = RunService(auth, conn_url)
        self.api = api



import re
from typing import Optional

from malevich_space.schema import LoadedComponentSchema

from malevich._core.scan import scan_core
from malevich._utility import Stub, get_auto_ops
from malevich.constants import DEFAULT_CORE_HOST
from malevich.manifest import ManifestManager
from malevich.models.installers.space import SpaceDependency, SpaceDependencyOptions
from malevich.path import Paths

from .installer import Installer

manf = ManifestManager()


class Templates:
    """Templates for creating a package from a Docker image"""

    disclaimer = """\"\"\"
THIS FILE IS AUTOGENERATED BY malevich PACKAGE INSTALLER.

THIS FILE CONTAINS VITAL INFORMATION ABOUT THE PACKAGE AND ITS CONTENTS.
DO NOT MODIFY THIS FILE MANUALLY.
\"\"\"
"""

    imports = """
from malevich._autoflow.function import autotrace, sinktrace
from malevich._utility.registry import Registry
from malevich.models.nodes import OperationNode

from pydantic import BaseModel
"""
    registry = """
Registry().register("{operation_id}", {{
    "reverse_id": "{reverse_id}",
    "branch": {branch},
    "version": {version},
    "processor_name": "{name}",
    "processor_id": "{name}",
    "image_ref": {image_ref},
    "image_auth_user": {image_auth_user},
    "image_auth_pass": {image_auth_pass},
}})
"""

    processor = """
@{decor}
def {name}({args}config: dict = {{}}):
    \"\"\"{docs}\"\"\"
    return OperationNode(
        operation_id="{operation_id}",
        config=config,
        processor_id="{name}",
        package_id="{package_id}",
    )
"""


class SpaceInstaller(Installer):
    name = 'space'

    def __init__(self) -> None:
        super().__init__()
        try:
            self.__ops = get_auto_ops()
        except Exception as e:
            from malevich._cli.space.login import login
            if not login():
                raise Exception(
                    "Login failed. Run `malevich space login` to try again"
                ) from e
            SpaceInstaller.__init__(self)


    def install(
        self,
        package_name: str,
        reverse_id: str,
        branch: Optional[str] = None,
        version: Optional[str] = None,
    ) -> SpaceDependency | LoadedComponentSchema:
        package_name = re.sub(r'[\W\s]+', '_', package_name)

        component = self.__ops.get_parsed_component_by_reverse_id(
            reverse_id=reverse_id
        )

        if component is None:
            raise Exception(f"Component {reverse_id} not found")

        if component.flow is not None:
            return component

        if component.app is None:
            raise Exception(f"Component {reverse_id} is not an app")


        m = ManifestManager()

        if component.app:
            iauth, itoken = m.put_secrets(
                image_auth_user=component.app.container_user,
                image_auth_password=component.app.container_token,
            )
        else:
            iauth, itoken = None, None

        dependency = SpaceDependency(
            package_id=package_name,
            version=component.version.readable_name,
            installer="space",
            options=SpaceDependencyOptions(
                reverse_id=reverse_id,
                branch=component.branch.uid,
                version=component.version.uid,
                image_ref=component.app.container_ref,
                image_auth_user=iauth,
                image_auth_pass=itoken
            )
        )

        Stub.from_app_info(
            app_info=scan_core(
                image_ref=component.app.container_ref,
                image_auth=(
                    component.app.container_user, component.app.container_token
                ),
                core_host=DEFAULT_CORE_HOST,
            ),
            path=Paths.module(package_name),
            package_name=package_name,
            dependency=dependency,
            operation_ids={
                op.core_id: op.uid
                for op in component.app.ops
            },
            registry_records={
                str(op.core_id): {
                    "operation_id": op.uid,
                    "reverse_id": reverse_id,
                    "branch": component.branch.uid,
                    "version": component.version.uid,
                    "processor_name": op.core_id,
                    "processor_id":  op.core_id,
                    "image_ref": ('dependencies', package_name, 'options', 'image_ref'),
                    "image_auth_user": (
                        'dependencies', package_name, 'options', 'image_auth_user'
                    ),
                    "image_auth_pass": (
                        'dependencies', package_name, 'options', 'image_auth_pass'
                    )
                }
                for op in component.app.ops
            },
            description=component.description
        )

        return dependency

    def restore(
            self,
            dependency: SpaceDependency
    ) -> SpaceDependency | LoadedComponentSchema:
        return self.install(
            package_name=dependency.package_id,
            reverse_id=dependency.options.reverse_id,
            branch=dependency.options.branch,
            version=dependency.options.version
        )

    def construct_dependency(self, dependency: dict) -> SpaceDependency:
        return SpaceDependency(**dependency)




from malevich._dev.singleton import SingletonMeta
from malevich._utility.package import package_manager as pm
from malevich.constants import TEST_DIR
from malevich.install.image import ImageInstaller
from malevich.install.installer import Installer
from malevich.install.space import SpaceInstaller
from malevich.manifest import OverrideManifest, manf
from malevich.models.dependency import Dependency
from malevich.models.installers.compat import CompatabilityStrategy
from malevich.models.preferences import (
    Action,
    LogFormat,
    UserPreferences,
    VerbosityLevel,
)

test_manifest = OverrideManifest(TEST_DIR)

class EnvManager(metaclass=SingletonMeta):
    """Efficient environment manager

    Provides optimized and flexible way for installing,
    tracking and removing environments for testing.
    """
    installers: dict[str, Installer] = {
        'image': ImageInstaller(),
        'space': SpaceInstaller()
    }
    """A set of know installers"""

    def __init__(
        self,
        compatability_strategy: CompatabilityStrategy = None,
    ) -> None:
        """Initialize new manager

        Args:
            compatability_strategy (CompatabilityStrategy):
                A stategy used to compare dependencies
        """
        if not compatability_strategy:
            compatability_strategy = CompatabilityStrategy(
                none_is_always_compatible=True
            )

        self._strategy = compatability_strategy
        self.clean_env()

    def clean_env(self) -> None:
        """Remove all installed stubs and dependencies"""
        with test_manifest:
            stubs, _ = self.get_current_env()
            manf.remove('dependencies')
            manf.put('dependencies', value=[])
            manf.put('preferences', value=UserPreferences(
                verbosity={
                    Action.Interpretation.value: VerbosityLevel.AllSteps.value,
                    Action.Preparation.value: VerbosityLevel.AllSteps.value,
                    Action.Run.value: VerbosityLevel.AllSteps.value,
                    Action.Stop.value: VerbosityLevel.AllSteps.value,
                    Action.Installation.value: VerbosityLevel.AllSteps.value,
                    Action.Removal.value: VerbosityLevel.AllSteps.value,
                    Action.Results.value: VerbosityLevel.AllSteps.value,
                },
                log_format = LogFormat.Plain,
                log_level="DEBUG"
            ))
            for stub in stubs:
                pm.remove_stub(stub[0])

    def get_stubs(self) -> list[tuple[str, str]]:
        """Retrieves a list of stubs

        Returns:
            tuple[str, str]: pairs (stub_name, stub_path)
        """
        all_ = pm.get_all_packages()
        return [(x, pm.get_package_path(x),) for x in all_]

    def get_manifested(self) -> list[Dependency]:
        """Retrieves a list of manifested dependencies

        Returns:
            list[Dependency]: a list of dependencies found in manifest
        """
        deps = []
        with test_manifest:
            for x in manf.query('dependencies'):
                obj = x[next(iter(x.keys()))]
                deps.append(self.installers[obj['installer']].construct_dependency(obj))
        return deps

    def get_current_env(self) -> tuple[list[str], list[Dependency]]:
        """Retrieves stubs and manifested dependencies"""
        stubs = self.get_stubs()
        manifested = self.get_manifested()
        return stubs, manifested

    def request_env(self, dependencies: list[Dependency]) -> None:
        """Provides a contract for a new environment"""
        _, manifested = self.get_current_env()
        for dependency in dependencies:
            # should_offload = False
            for m in manifested:
                if (
                    dependency.package_id == m.package_id
                    and dependency.compatible_with(m, self._strategy)
                ):
                    break
            else:
                with test_manifest:
                    try:
                        pm.remove_stub(dependency.package_id)
                    except Exception:
                        pass
                    manf.remove('dependencies', dependency.package_id)
                    restored = (
                        self.installers[dependency.installer].restore(dependency).model_dump()
                    )

                    manf.put(
                        'dependencies',
                        dependency.package_id,
                        value=restored,
                    )

        return self.get_current_env()



from __future__ import annotations

import asyncio
import enum
from abc import ABC, abstractmethod
from typing import Any, Callable, Generic, ParamSpec, TypeVar

from malevich.models import BaseInjectable, MetaEndpoint, Result, TreeNode
from malevich.types import FlowOutput

State = TypeVar("State")
CallbackParams = ParamSpec('CallbackParams')
Callback = Callable[CallbackParams, None]
StageClass = TypeVar("StageClass", bound=enum.Enum)


class BaseTask(ABC, Generic[StageClass]):
    supports_conditional_output: bool = False

    @property
    @abstractmethod
    def tree(self) -> TreeNode:
        """Returns the execution tree, encapsulated into TreeNode class"""
        pass

    @abstractmethod
    def get_stage(self) -> StageClass:
        pass

    def get_stage_class(self) -> type[StageClass]:
        return StageClass

    # Main Task Operations
    # ====================
    @abstractmethod
    def interpret(self, interpreter: Any = None) -> None:  # noqa: ANN401
        """Attaches the task to particular platform"""
        pass

    @abstractmethod
    async def prepare(self, *args, **kwargs) -> None:
        """Prepares the task for execution"""
        pass

    @abstractmethod
    async def run(self, run_id: str | None = None, *args, **kwargs) -> None:
        """Performs a single run with a given ID"""
        pass

    @abstractmethod
    async def stop(self, *args, **kwargs) -> None:
        """Stops the task preventing it from further execution"""
        pass

    @abstractmethod
    async def results(self, *args, **kwargs) -> list[Result]:
        """Retrieves results of the run"""
        pass

    # Auxillary methods (used by flows)
    # =================================
    @abstractmethod
    def commit_returned(self, returned: FlowOutput) -> None:
        """Reports what part of the execution tree should be used to retrieve results"""
        pass

    # Saving / Loading
    # ================
    @abstractmethod
    def dump(self) -> bytes:
        """Serializes the task into bytes"""
        pass

    @staticmethod
    @abstractmethod
    def load(self, object_bytes: bytes) -> BaseTask:
        """Deserializes the task from bytes"""
        pass

    # Configuration methods (for runs)
    # ================================
    @abstractmethod
    def get_injectables(self) -> list[BaseInjectable]:
        """Returns a list of possible tweaks for the run"""
        pass

    @abstractmethod
    def get_operations(self) -> list[str]:
        """Returns a operations within the task"""
        pass

    @abstractmethod
    def configure(
        self,
        *operations: str,
        # Configurable parameters
        platform: str = 'base',
        platform_settings: dict[str, Any] | None = None,
        **kwargs
    )-> None:
        """Configures a given operation within the task"""
        pass

    # Interpretation
    # ==============
    @abstractmethod
    def get_interpreted_task(self) -> BaseTask:
        """Returns the task interpreted for a particular platform"""
        pass


    def publish(self, *args, **kwargs) -> MetaEndpoint:
        raise NotImplementedError(
            f"Publishing is not available for {self.__class__.__name__}"
        )




import json
import os
from pathlib import Path
from typing import Literal, Type, TypeVar, overload

import pandas as pd
from pydantic import BaseModel

from malevich.models.results.base import BaseResult
from malevich.path import Paths

Types = Literal['df', 'doc',  'obj']
DocumentType = TypeVar('DocumentType', bound=BaseModel)

class LocalResult(BaseResult):
    def __init__(
        self,
        operation_id: str,
        run_id: str,
        name: str,
        login: str,
    ) -> None:
        self.__path = Path(Paths.local_results(
            operation_id,
            run_id,
            name
        ))
        self.__login = login
        self.__type = (
            'df' if str(next(self.__path.iterdir())).endswith('.csv') else 'doc'
        ) if not self._is_asset() else 'obj'

    def __get_mnt_path(self, path) -> str:
        return Paths.local(
            'mnt_obj',
            self.__login,
            path
        )

    def _is_asset(self) -> bool:
        return (
            str(self.__path).endswith('.csv')
            and (df := pd.read_csv(self.__path)).columns == ['path']
            and all(
                os.path.exists(self.__get_mnt_path(path))
                for path in df['path']
            )
        )

    def get_df(self) -> pd.DataFrame:
        if self.__type != 'df':
            raise TypeError("Result is not a dataframe")
        return pd.read_csv(
            self.__path / '0.csv'
        )

    def get_dfs(self) -> list[pd.DataFrame]:
        return [
            pd.read_csv(file)
            for file in self.__path.glob('*.csv')
        ]

    @overload
    def get_document(self, index: int = 0) -> dict:
        """Returns dumped document from the result if possible

        Returns:
            dict: The dumped document
        """
        pass

    @overload
    def get_document(
        self,
        index: int = 0,
        *,
        model: Type[DocumentType] | None = None
    ) -> DocumentType:
        """Returns document from the result if possible

        Returns:
            DocumentType: The document
        """
        pass

    def get_document(
        self,
        index: int = 0,
        model: Type[DocumentType] | None = None
    ) -> dict | DocumentType:
        if self.__type != 'doc':
            raise TypeError("Result is not a document")

        if model is not None:
            return model.model_validate_json(
                open(self.__path / f'{index}.json').read()
            )
        return json.loads(open(self.__path / f'{index}.json').read())

    def get_documents(self) -> list[dict]:
        return [
            json.loads(file.read_text())
            for file in self.__path.glob('*.json')
        ]

    def get_binary(self) -> bytes:
        if not self._is_asset():
            raise TypeError(f"Result is not an asset, but {self.__type}")
        return open(self.__get_mnt_path(
            self.get_df().iloc[0]['path']
        ), 'rb').read()

    def get_binary_dir(self) -> dict[str, bytes]:
        if not self._is_asset():
            raise TypeError(f"Result is not an asset, but {self.__type}")
        return {
            file.name: open(self.__get_mnt_path(file.name), 'rb').read()
            for file in self.__path.glob('*')
        }

    def get(self) -> list[pd.DataFrame]:
        return self.get_dfs()

    @property
    def num_elements(self) -> int:
        return len(os.listdir(self.__path))



from typing import Any, Optional

from malevich_space.schema import SpaceSetup
from pydantic import BaseModel

from .dependency import Dependency, Integration
from .preferences import UserPreferences


class Manifest(BaseModel):
    project_id: Optional[str] = None
    version: Optional[str] = None
    space: Optional[SpaceSetup] = None
    dependencies: list[dict[str, Dependency]] = []
    preferences: UserPreferences = UserPreferences()
    flows: dict[str, list[Integration]] = {}


class ManifestUpdateEntry(BaseModel):
    package_id: str
    version: str
    installer: str
    options: dict[str, Any]


class Secret(BaseModel):
    secret_key: str
    secret_value: str
    salt: Optional[str] = None
    external: Optional[bool] = False


class Secrets(BaseModel):
    secrets: dict[str, Secret] = {}



import sys
from datetime import datetime

from malevich._db import Write

from .._db.schema.artifacts import MasterSession, Session
from .._utility.malevich_version import get_malevich_version


def decide_session() -> Session:
    with Write() as db_ref:
        sessions = db_ref.query(Session).all()

        # Search for session less than an hour ago
        current_time = datetime.now()
        sessions = [
            session for session in sessions if session.end is not None
        ]

        sessions.sort(key=lambda x: x.end, reverse=True)
        r_session = None
        for session in sessions:
            if session.end and (current_time - session.end).seconds < 3600:
                r_session = session
                break

        if r_session is None:
            m_session = MasterSession()
            db_ref.add(m_session)
            db_ref.commit()
            r_session = Session(
                start=current_time,
                python_version='.'.join(map(str, sys.version_info[:3])),
                malevich_version=get_malevich_version(),
                master_id=m_session.id
            )
            db_ref.add(r_session)
            db_ref.commit()

        return r_session.id



import hashlib
from typing import Optional

from pydantic import BaseModel

from malevich.models.dependency import Dependency

from .compat import CompatabilityStrategy, compare_images


class SpaceDependencyOptions(BaseModel):
    reverse_id: str
    branch: Optional[str] = None
    version: Optional[str] = None
    image_ref: Optional[str] = None
    image_auth_user: Optional[str] = None
    image_auth_pass: Optional[str] = None


class SpaceDependency(Dependency):
    installer: str = 'space'
    options: SpaceDependencyOptions

    def checksum(self) -> str:
        return hashlib.sha256(self.model_dump_json().encode()).hexdigest()

    def compatible_with(
        self, other: Dependency, compatability_strategy: CompatabilityStrategy
    ) -> bool:
        if isinstance(other, SpaceDependency):
            compat = self.options.reverse_id == other.options.reverse_id
            if self.options.branch is not None:
                compat &= self.options.branch == other.options.branch
            if self.options.version is not None:
                compat &= self.options.version == other.options.version
            if self.options.image_ref is not None:
                compat &= compare_images(
                    self.options.image_ref,
                    other.options.image_ref,
                    compatability_strategy
                )
            return compat
        return False



from pydantic import BaseModel


class BaseAction(BaseModel):
    pass


class CLIAction(BaseAction):
    full_command: str

class CoreRequestAction(BaseAction):
    query: str
    response_status: int
    response_raw: bytes
    response_media_type: str



import os
import sys
from typing import Any, Type

from pydantic import BaseModel
from pytest import fixture

from malevich.constants import TEST_DIR
from malevich.interpreter.abstract import Interpreter
from malevich.manifest import OverrideManifest
from malevich.models.flow_function import FlowFunction
from malevich.models.installers.image import ImageDependency
from malevich.models.installers.space import SpaceDependency
from malevich.models.results.base import BaseResult
from malevich.models.task.interpreted.space import SpaceTask
from malevich.models.task.promised import PromisedTask

from .env import EnvManager

env_manager = EnvManager()
manifest_override = OverrideManifest(TEST_DIR)

@fixture(autouse=True, scope='session')
def clean_env() -> None:
    global env_manager
    # env_manager.clean_env()

class FlowTestEnv(BaseModel):
    dependencies: dict[str, ImageDependency | SpaceDependency] = []
    env_vars: dict[str, str] = {}


class FlowTestSuite:
    environment: FlowTestEnv = None
    func_args: list[Any] = []
    func_kwargs: dict[str, Any] = {}
    prepare_args: list[Any] = []
    prepare_kwargs: dict[str, Any] = {}
    run_args: list[Any] = []
    run_kwargs: dict[str, Any] = {}
    result_kwargs: dict[str, Any] = {}
    interpreter: Interpreter = None

    @staticmethod
    def __offload_modules(stubs) -> None:
        packages = ['malevich.' + x[0] for x in stubs]
        for package in packages:
            for module in [*sys.modules.keys()]:
                if package in module:
                    sys.modules.pop(module)


    @staticmethod
    def on_interpretation(task: PromisedTask) -> None:
        pass

    @staticmethod
    def on_interpretation_error(task: PromisedTask, error: Exception) -> Exception | None:  # noqa: E501
        return error

    @staticmethod
    def on_prepare_start(flow: FlowFunction) -> None:
        pass

    @staticmethod
    def on_prepare_end(flow: FlowFunction, output: Any) -> None:  # noqa: ANN401
        pass

    @staticmethod
    def on_prepare_error(flow: FlowFunction, error: Exception) -> Exception | None:
        return error

    @staticmethod
    def on_run_start(flow: FlowFunction) -> None:
        pass

    @staticmethod
    def on_run_end(flow: FlowFunction, output: Any) -> None:  # noqa: ANN401
        pass

    @staticmethod
    def on_run_error(flow: FlowFunction, error: Exception) -> Exception | None:
        return error

    @staticmethod
    def on_result(flow: FlowFunction, results: list[BaseResult]) -> None:
        pass

    @staticmethod
    def on_result_error(flow: FlowFunction, error: Exception) -> Exception | None:
        return error

    @classmethod
    def test_flow(cls: Type['FlowTestSuite']) -> None:
        old_stubs, _ = env_manager.current_env()
        cls.__offload_modules(old_stubs)
        env_manager.request_env([*cls.environment.dependencies.values()])
        stubs, _ = env_manager.current_env()
        for k, v in cls.environment.env_vars.items():
            os.environ[k] = v
        cls.__offload_modules(stubs)
        for attr in dir(cls):
            f = getattr(cls, attr)
            if isinstance(f, FlowFunction):
                with manifest_override:
                    task: PromisedTask = f(
                        *cls.func_args,
                        **cls.func_kwargs
                    )
                    try:
                        task.interpret(cls.interpreter)
                        if isinstance(t := task.get_interpreted_task(), SpaceTask):
                            # FIXME: local fix for upload, remove later
                            print("UPLOADING")
                            t.upload()
                        cls.on_interpretation(task)
                    except Exception as e:
                        if e_ := cls.on_interpretation_error(task, e):
                            raise e_
                    if not isinstance(task, PromisedTask):
                        raise TypeError(
                            f"FlowFunction {f} did not return a PromisedTask"
                        )

                    try:
                        cls.on_prepare_start(f)
                        cls.on_prepare_end(
                            f,
                            task.prepare(
                                *cls.prepare_args,
                                **cls.prepare_kwargs
                            )
                        )
                    except Exception as e:
                        if e_ := cls.on_prepare_error(f, e):
                            raise e_


                    try:
                        cls.on_run_start(f)
                        cls.on_run_end(
                            f,
                            task.run(
                                *cls.run_args,
                                **cls.run_kwargs
                            )
                        )
                    except Exception as e:
                        if e_ := cls.on_run_error(f, e):
                            raise e_

                    # try:
                    #     task
                    # except Exception:
                    #     pass

                    try:
                        results = task.results(
                            **cls.result_kwargs
                        )
                        cls.on_result(f, results)
                    except Exception as e:
                        if e_ := cls.on_result_error(f, e):
                            raise e_

        for k in cls.environment.env_vars:
            os.environ.pop(k)



from pydantic import BaseModel

from malevich.core_api import ResultDoc


class SpaceCallbackResult(BaseModel):
    in_flow_id: str
    core_id: str
    docs: list[ResultDoc]
    in_flow_alias: str | None = None
    core_alias: str | None = None
    core_username: str | None = None
    core_password: str | None = None


class SpaceCallbackBody(BaseModel):
    run_id: str | None = None
    task_id: str | None = None
    in_flow_id: str | None = None
    results: list[SpaceCallbackResult]





from malevich_space.schema import SpaceSetup


def resolve_setup(space: dict) -> SpaceSetup:
    from malevich.manifest import manf
    space = SpaceSetup(**space)

    if manf.is_secret(space.password):
        space.password = manf.query_secret(space.password, only_value=True)

    return space



import json
import os
import logging
from datetime import datetime
from pydantic import BaseModel
from malevich._dev.singleton import SingletonMeta
from malevich.path import Paths
from malevich.models.installers.compat import CompatabilityStrategy
from malevich.install.installer import Installer
from malevich.manifest import ManifestManager
from malevich.models.dependency import Dependency
from malevich.constants import TEST_DIR
from malevich.install.image import ImageInstaller
from malevich.install.space import SpaceInstaller
from malevich._utility import CacheManager, package_manager

env_manf = ManifestManager(TEST_DIR)
cache = CacheManager()

class _OffloadedDependency(BaseModel):
    """When dependency is offloaded, it is saved to this model"""
    orig_stub_path: str
    offloaded_stub_path: str
    dependency: Dependency | None = None

class EnvManager(metaclass=SingletonMeta):
    session = datetime.now().strftime('%d.%m.%y__%H:%M:%S')
    installers: dict[str, Installer] = {
        'image': ImageInstaller(),
        'space': SpaceInstaller()
    }
    def _setup_logger(self) -> None:
        self._logger = logging.getLogger('malevich.testing')
        self._logger.handlers.clear()
        formatter = logging.Formatter(
            "%(asctime)s -- %(funcName)s:%(lineno)d -- %(message)s"
        )
        handler = logging.FileHandler(
            cache.testing.probe_new_entry('protocol.log', entry_group=self.session)[-1],
            mode='a+'
        )
        handler.setFormatter(formatter)
        self._logger.addHandler(handler)

    def __init__(self) -> None:
        self._offloaded: list[_OffloadedDependency] = []
        self._logger = None
        self._setup_logger()


    def offload_stub(  # noqa: ANN201
        self,
        stub_id: str,
        dont_fail: bool = True
    ):
        stub_path = package_manager.get_package_path(stub_id)
        if os.path.exists(stub_path):
            off_path = cache.testing.cache_dir(
                stub_path, entry_group='offload'
            )
            return _OffloadedDependency(
                orig_stub_path=stub_path,
                offloaded_stub_path=off_path,
            )
        elif not dont_fail:
            raise Exception(f"No stub found for {stub_id}")

    def offload_manf(
        self,
        package_name: str,
        dont_fail: bool = False
    ) -> None:
        try:
            env_manf.remove('dependencies', package_name)
        except Exception:
            if not dont_fail:
                raise



    def current_env(self) -> tuple[dict[str, Dependency], list[str]]:
        manifest_keys = [
            next(iter(key.keys()))
            for key in env_manf.query('dependencies')
        ]
        manifested_raw = [
            env_manf.query('dependencies', key)
            for key in manifest_keys
        ]
        manifested: dict[str, Dependency] = {
            key: (
                self.installers[x['installer']]
                .construct_dependency(x)
            )
            for key, x in zip(
                manifest_keys,
                manifested_raw
            )
        }
        stubs = package_manager.get_all_packages()
        return manifested, stubs

    def install(self, dependency: Dependency) -> None:
        installer = dependency.installer
        dependency = self.installers[installer].restore(dependency)
        env_manf.put(
            'dependencies',
            dependency.package_id,
            value={dependency.package_id: dependency.model_dump()},
            append=True
        )


    def request_env(
        self,
        dependencies: list[Dependency | tuple[str, Dependency]]
    ) -> None:
        manifested, _ = self.current_env()
        for record in dependencies:
            if isinstance(record, tuple):
                package_name, dependency = record
            else:
                package_name = record.package_id
                dependency = record

            package_id = dependency.package_id
            to_offload_ = None
            for manifest_name, manifest_dependency in manifested.items():
                if dependency.compatible_with(
                    manifest_dependency,
                    compatability_strategy=CompatabilityStrategy()
                ):
                    break
                if manifest_name == package_name:
                    # Found collision in manifest: should be offloaded
                    to_offload_ = manifest_name, manifest_dependency
            else:
                # If no compatible dependency is found, offload the stub
                off_ = self.offload_stub(package_id)
                if to_offload_:
                    off_.dependency = to_offload_[-1]
                    self.offload_manf(to_offload_[0])
                self._offloaded.append(off_)

                # And install our dependency
                self.install(dependency)
                self._logger.info(
                    f"install {package_id}, dependency= "
                    + json.dumps(dependency.model_dump(), separators=(',', ':'))
                )



import hashlib
import os
import shutil
import sys

from .._dev.singleton import SingletonMeta


class PackageManager(metaclass=SingletonMeta):
    builtins = [
        "_autoflow",
        "_core",
        "_meta",
        "_templates",
        "_utility",
        "commands",
        "install",
        "interpreter",
        "models",
        "square",
        "runners"
    ]

    def __init__(self) -> None:
        self.__malevich_path = os.path.dirname(
            sys.modules["malevich"].__file__
        )

    def is_builtin(self, package_name: str) -> bool:
        return package_name in self.builtins

    def get_malevich_path(self) -> str:
        return self.__malevich_path

    def get_package_path(self, package_name: str) -> str:
        return os.path.join(self.__malevich_path, package_name)

    def create_stub(self, package_name: str, package_content: str) -> None:
        package_path = self.get_package_path(package_name)
        if os.path.exists(package_path):
            raise Exception(f"Package {package_name} already exists")
        os.mkdir(package_path)
        with open(os.path.join(package_path, "__init__.py"), "w") as f:
            f.write(package_content)

    def remove_stub(self, package_name: str) -> None:
        package_path = self.get_package_path(package_name)
        if not os.path.exists(package_path):
            raise Exception(f"Package {package_name} does not exist")
        shutil.rmtree(package_path)

    def get_package(self, package_name: str) -> str:
        package_path = self.get_package_path(package_name)
        if not self.is_package(package_path):
            raise Exception(f"Package {package_name} does not exist")
        return package_path

    def is_package(self, package_fol_path: str) -> bool:
        is_package = os.path.exists(package_fol_path)
        is_package &= os.path.isdir(package_fol_path)
        init_py = os.path.join(package_fol_path, '__init__.py')
        is_package &= os.path.exists(init_py)
        is_package &= os.path.isfile(init_py)
        is_old_package = False
        is_new_package = False

        F_py = os.path.join(package_fol_path, 'F.py')  # noqa: N806
        scheme_py = os.path.join(package_fol_path, 'scheme.py')
        if os.path.exists(F_py) and os.path.exists(scheme_py):
            is_new_package = True

        if is_package:
            is_old_package = is_package
            with open(init_py) as f:
                str_ = ""
                for line in f.readlines():
                    if line.startswith('__Metascript_checksum__'):
                        exec(line)
                        checksum = locals().get('__Metascript_checksum__')
                        n_checksum =  hashlib.sha256(str_[:-1].encode()).hexdigest()
                        is_old_package &= checksum == n_checksum
                        break
                    else:
                        str_ += line
                else:
                    is_old_package = False
        return is_new_package or is_old_package

    def get_all_packages(self) -> list[str]:
        return [
            package
            for package in os.listdir(self.__malevich_path)
            if self.is_package(self.get_package_path(package))
        ]

    def is_importable(self, package_name: str) -> bool:
        import importlib
        try:
            lib = importlib.import_module(f"malevich.{package_name}")
        except Exception:
            return False
        return lib is not None


package_manager = PackageManager()



from malevich_space.schema import LoadedComponentSchema, LoadedFlowSchema


def get_space_leaves(element: LoadedComponentSchema | LoadedFlowSchema):
    if isinstance(element, LoadedComponentSchema):
        flow = element.flow
    else:
        flow = element

    if flow is None:
        raise ValueError("Flow is not loaded")

    memory_ = set()
    for node in flow.components:
        if node.prev:
            for prev in node.prev:
                memory_.add(prev.uid)

    return [
        node for node in flow.components if node.uid not in memory_
    ]



from functools import partial, wraps
from inspect import Parameter, signature

from typer.models import CommandInfo


def make_command(core_op, *args, callback=None, **kwargs) -> CommandInfo:
    fn = partial(core_op, auth=None, conn_url=None, batcher=None)

    @wraps(callback)
    def wrapped(*args, **kwargs) -> None:
        results = fn(*args, **kwargs)
        if callback:
            callback(results)
        else:
            print(results)

    # set wrapped args and kwargs to
    # fn's signature
    annotations = fn.func.__annotations__
    for k in fn.keywords:
        annotations.pop(k, None)
    annotations.pop('return', None)
    sig = signature(fn)
    sig = sig.replace(parameters=[
        Parameter(name=k, kind=Parameter.POSITIONAL_OR_KEYWORD, annotation=v)
        for k, v in annotations.items()
    ])
    wrapped.__annotations__ = annotations
    wrapped.__signature__ = sig
    command = CommandInfo(
        *args,
        callback=wrapped,
        **kwargs,
    )
    return command


def wrap_command(core_f, exclude: list[str] | None = None):
    if not exclude:
        exclude = []

    fn = partial(core_f, auth=None, conn_url=None, batcher=None)

    def decorator(f):
        @wraps(f)
        def wrapped(*args, **kwargs) -> None:
            return f(*args, **kwargs)

        # set wrapped args and kwargs to
        # fn's signature
        annotations = fn.func.__annotations__
        for k in fn.keywords:
            annotations.pop(k, None)
        annotations.pop('return', None)
        fn_sig = signature(fn)
        fn_sig = fn_sig.replace(parameters=[
            x for x in fn_sig.parameters.values()
            if x.name in annotations
        ])
        f_signature = signature(f)
        # merge annotations
        wrapped.__annotations__ = {**annotations, **f.__annotations__}
        wrapped.__annotations__.pop('kwargs', None)

        params = [
            *fn_sig.parameters.values(), *f_signature.parameters.values()
        ]
        names = {'kwargs', *exclude}
        unique_params: list[Parameter] = []
        for p in params:
            if p.name not in names:
                unique_params.append(p)
            names.add(p.name)
        unique_params.sort(key=lambda x: x.default != Parameter.empty)
        # merge signatures
        f_signature = f_signature.replace(
            parameters=unique_params
        )

        wrapped.__signature__ = f_signature
        return wrapped

    return decorator



import os
import shutil
import uuid

from malevich._dev.singleton import SingletonMeta


class CacheController:
    def __init__(
        self,
        user_cache_path: os.PathLike,
        catalog_path: os.PathLike
    ) -> None:
        self._user_cache_path = user_cache_path
        self._cat_path = catalog_path

    @property
    def path(self) -> str:
        return os.path.join(self._user_cache_path, self._cat_path)

    def get_entry_path(
        self,
        entry_name: str | None = None,
        entry_group: os.PathLike | None = None,
    ) -> str:
        """Constructs the path to a new cache entry

        Args:
            - entry_name: The name of cache entity (file).
                Autogenerated by default.
            - entity_group: Path-like string to denote path to
                the entity within the cache catalog
        """
        if not entry_name:
            entry_name = uuid.uuid4().hex

        if entry_group:
            return os.path.join(
                self.path,
                entry_group,
                entry_name
            )
        else:
            return os.path.join(
                self.path,
                entry_name
            )


    def probe_new_entry(
        self,
        entry_name: str | None = None,
        entry_group: str | None = None,
    ) -> tuple[bool, str]:
        """Ensures the entry under specified name and group can be created

        Args:
            - `entry_name (str, optional)`: The name of cache entity (file).
                Autogenerated by default.
            - `entity_group (str, optional)`: Path-like string to denote path to
                the entity within the cache catalog

        Returns:
            - `ok (bool)` - whether the entry is free
            - `path (str)` - a path to an entry
        """
        path_ = self.get_entry_path(
            entry_name,
            entry_group,
        )

        if os.path.exists(path_):
            return False, path_

        os.makedirs(
            os.path.dirname(path_),
            exist_ok=True
        )

        return True, path_


    def write_entry(
        self,
        data: str | bytes,
        *,
        entry_name: str | None = None,
        entry_group: os.PathLike | None = None,
        force_overwrite: bool = False,
        append: bool = False,
    ) -> bool:
        """Writes to the cache entry

        Args:
            - data (str or bytes): Text or byte data to write
            - entry_name (str, optional): Name of the entry.
                Autogenerated by default.
            - entry_group (str, optional): Path-like name of group.
            - force_overwrite (bool, optional): If true, overwrite entry if exists
        """
        ok_, path_ = self.probe_new_entry(
            entry_name,
            entry_group
        )
        if not ok_ and not force_overwrite:
            raise Exception(
                "Cache entry exists. "
                f"cat: {self._cat_path}, "
                f"entry g.: {entry_group or '!not given'}, "
                f"entry n.: {entry_name}, "
                f"entry full path: {path_}"
            )

        if isinstance(data, bytes):
            if append:
                mode = 'ab+'
            else:
                mode = 'wb+'
        elif isinstance(data, str):
            if append:
                mode = 'a+'
            else:
                mode = 'w+'
        else:
            raise TypeError(
                "expected arg. `data` to be either bytes or str, "
                f" but got {type(data)}"
            )

        with open(
            path_,
            mode=mode
        ) as file_:
            file_.write(data)

        return path_

    def cache_file(
        self,
        path: os.PathLike,
        entry_name: str | None = None,
        entry_group: os.PathLike | None = None,
    ) -> os.PathLike:
        """Caches entire file

        Args:
            - path (str): Path to the file to cache
            - entry_name (str, optional): Name of the entry.
                Autogenerated by default.
            - entry_group (str, optional): Path-like name of group.

        Returns:
            - path (str): Path to the cached file
        """
        ok_, path_ = self.probe_new_entry(
            entry_name=entry_name,
            entry_group=entry_group
        )

        if not ok_:
            raise ValueError(
                f"There exists a record for entry_name={entry_name} "
                f"and entry_group={entry_group}"
            )

        shutil.move(
            path,
            path_
        )

        return path_

    def cache_dir(
        self,
        path: os.PathLike,
        entry_name: str | None = None,
        entry_group: os.PathLike | None = None
    ) -> os.PathLike:
        """Caches entire directory

        Args:
            - path (str): Path to the directory to cache
            - entry_name (str, optional): Name of the entry.
                Autogenerated by default.
            - entry_group (str, optional): Path-like name of group.

        Returns:
            - path (str): Path to the cached directory
        """
        ok_, path_ = self.probe_new_entry(
            entry_name=entry_name,
            entry_group=entry_group
        )

        if not ok_:
            raise ValueError(
                f"There exists a record for entry_name={entry_name} "
                f"and entry_group={entry_group}"
            )

        shutil.move(
            path,
            path_
        )

        return path_

    def is_dir(
        self,
        entry_name: str | None = None,
        entry_group: os.PathLike | None = None
    ) -> bool:
        """Check whether the entry is a directory"""
        pth = self.get_entry_path(
            entry_name=entry_name,
            entry_group=entry_group
        )
        return os.path.isdir(pth)

    def is_file(
        self,
        entry_name: str | None = None,
        entry_group: str | None = None,
    ) -> bool:
        """Check whether the entry is a file"""
        pth = self.get_entry_path(
            entry_name=entry_name,
            entry_group=entry_group
        )
        return os.path.isfile(pth)


class CacheManager(metaclass=SingletonMeta):
    def __init__(self) -> None:
        self._fs = os.path.expanduser(
            os.getenv("MALEVICH_CACHE_PATH", '~/.malevich/cache')
        )

    def __getattr__(self, __name: str) -> CacheController:
        if __name not in dir(self):
            return CacheController(self._fs, __name)
        else:
            object.__getattribute__(self, __name)



from .table import table
from .models import *
from .interpreter import *
from ._meta import *
from ._meta.asset import AssetFactory as asset  # noqa: N813
import malevich_coretools as core_api


# Manifest
from .manifest import *

# Expose contants
from .constants import *


# Malevich Core
from malevich_space.schema import VersionMode

# For simple deployment
from ._deploy import Core, Space, Local

# Handy ops
from ._utility import get_auto_ops

from malevich.models.overrides import CollectionOverride as __c
__c.model_rebuild()



"""
This module contains safe and fast version of
batched operations on Malevich Core

"""
import os
from concurrent.futures import Future, ProcessPoolExecutor
from multiprocessing import cpu_count
from typing import Optional

import malevich_coretools as core
from malevich_coretools import FilesDirs

from malevich.constants import DEFAULT_CORE_HOST
from malevich.models import AssetNode, Collection

executor = ProcessPoolExecutor(max_workers=cpu_count())


def result_collection_name(operation_id: str, alias: str = '') -> str:
    return f"result-{operation_id}-{alias}"


def _create_app_safe(
        app_id: str,
        extra: dict,
        uid: str,
        auth: core.AUTH = None,
        conn_url: Optional[str] = DEFAULT_CORE_HOST,
        *args,
        alias: str = '',
        **kwargs,
) -> None:
    settings = core.AppSettings(
        appId=app_id,
        taskId=app_id,
        saveCollectionsName=result_collection_name(uid, alias)
    )

    kwargs_ = {
        "app_id": app_id,
        "app_cfg": kwargs['app_cfg'],
        "image_ref": kwargs['image_ref'],
        "extra_collections_from": kwargs['extra_collections_from'],
    }

    try:
        core.create_app(
            app_id,
            processor_id=extra["processor_id"],
            auth=auth,
            conn_url=conn_url,
            *args,
            **kwargs
        )
    except Exception:
        pass
    else:
        return settings, kwargs_

    try:
        if core.get_app(
            app_id,
            auth=auth,
            conn_url=conn_url,
        ):
            core.delete_app(
                app_id,
                auth=auth,
                conn_url=conn_url,
            )
            core.delete_task(
                app_id,
                auth=auth,
                conn_url=conn_url,
            )
    except Exception:
        pass

    try:
        core.create_app(
            app_id,
            processor_id=extra["processor_id"],
            auth=auth,
            conn_url=conn_url,
            *args,
            **kwargs
        )
    except Exception as e:
        if 'processor_id' in extra:
            processor_id = extra['processor_id']
            raise Exception(
                f"Failed to create an app. Processor is {processor_id}. "
            ) from e
        else:
            raise Exception(
                "Failed to create an app and could determine the processor. "
                "Most probably, the app is installed incorrectly. Use "
                "malevich remove to remove it and reinstall it correctly"
            ) from e

    return settings, kwargs


def batch_create_apps(
    kwargs_list: list[dict],
) -> list[tuple[core.AppSettings, dict]]:
    results: list[Future] = []
    for kwargs_ in kwargs_list:
        results.append(executor.submit(_create_app_safe, **kwargs_))

    return [r.result() for r in results]


def _create_task_safe(
    auth: core.AUTH = None,
    conn_url: Optional[str] = DEFAULT_CORE_HOST,
    **task_kwargs,
) -> None:
    core.create_task(
        **task_kwargs,
        auth=auth,
        conn_url=conn_url,
    )


def batch_create_tasks(
    kwargs_list: list[dict],
    auth: core.AUTH = None,
    conn_url: Optional[str] = DEFAULT_CORE_HOST,
) -> None:
    # results: list[Future] = []
    # for kwargs_ in kwargs_list:
    #     results.append(executor.submit(_create_task_safe, **kwargs_))

    # return [r.result() for r in results]
    with core.Batcher(auth=auth, conn_url=conn_url):
        for kwargs_ in kwargs_list:
           _create_task_safe(**kwargs_)


def _upload_collection(
    collection: Collection,
    auth: core.AUTH = None,
    conn_url: Optional[str] = None,
) -> str:
    if collection.collection_data is None:
        raise Exception(
            f"Trying to upload collection {collection.collection_id} "
            "without data. Probably you set persistent=True and core_id, "
            "but collection was not found in Core."
        )

    if collection.collection_data is not None:
        collection.core_id = core.create_collection_from_df(
            data=collection.collection_data,
            name=collection.magic(),
            auth=auth,
            conn_url=conn_url,
        )

    return collection.core_id


def _assure_collection(
    collection: Collection,
    auth: core.AUTH = None,
    conn_url: Optional[str] = None,
) -> str:
    if collection.core_id:
        try:
            core.get_collection(collection.core_id, auth=auth, conn_url=conn_url)
            return collection.core_id
        except Exception as e:
            raise Exception(
                f"Collection {collection.collection_id} with core_id "
                f"{collection.core_id} is not found in Core. Do not "
                "specify core_id if you want to upload the collection."
            ) from e

    _ids = core.get_collections_by_name(
        collection.magic(),
        conn_url=conn_url,
        auth=auth,
    )

    if len(_ids.ownIds) == 0:
        _upload_collection(collection, auth, conn_url)
    else:
        collection.core_id = _ids.ownIds[0]

    # if collection.core_id not in core.get_collections(
    #     conn_url=conn_url,
    #     auth=auth
    # ).ownIds:
    #     raise Exception(
    #         f"Collection {collection.collection_id} with core_id "
    #         f"{collection.core_id} is not found in Core."
    #     )

    return collection.core_id


def batch_upload_collections(
    collections: list[Collection],
    auth: core.AUTH = None,
    conn_url: Optional[str] = None,
) -> list[str]:
    results: list[Future] = []
    for collection in collections:
        results.append(executor.submit(_assure_collection, collection, auth, conn_url))

    results = [r.result() for r in results]

    assert len(set(results)) == len(results), \
        "Some collections have the same core_id"

    return results

def _join_path(_path: str, _name: str) -> str:
    if _path.endswith("/"):
        return _path + _name
    else:
        return _path + "/" + _name

def _upload_asset(
    asset: AssetNode,
    auth: core.AUTH = None,
    conn_url: Optional[str] = None,
) -> None:
    if asset.is_composite:
        for f_ in asset.real_path:
            core.update_collection_object(
                _join_path(asset.core_path, os.path.basename(f_)),
                data=open(f_, "rb").read(),
                auth=auth,
                conn_url=conn_url,
            )
    else:
        core.update_collection_object(
            asset.core_path,
            data=open(asset.real_path, "rb").read(),
            auth=auth,
            conn_url=conn_url,
        )

def _assure_asset(
    asset: AssetNode,
    auth: core.AUTH = None,
    conn_url: Optional[str] = None,
) -> None:
    try:
        if not asset.is_composite:
            core.get_collection_object(
                asset.core_path,
                auth=auth,
                conn_url=conn_url,
            )
            objs = FilesDirs(files={asset.core_path: 0}, directories=[])
        else:
            objs = core.get_collection_objects(
                asset.core_path,
                auth=auth,
                conn_url=conn_url,
            )
    except Exception as e:
        if asset.real_path is not None:
            _upload_asset(asset, auth, conn_url)
            return
        else:
            raise Exception(
                f"Asset {asset.core_path} is not found in Core. Cannot "
                "upload it because real_path is not specified."
            ) from e

    if asset.real_path is not None:
        if asset.is_composite:
            for f_ in asset.real_path:
                if os.path.basename(f_) not in objs.files:
                    _upload_asset(asset, auth, conn_url)
                    break
        else:
            if asset.core_path not in objs.files:
                _upload_asset(asset, auth, conn_url)



from typing import Any, Generic, TypeVar

from pydantic import ConfigDict

from malevich._autoflow.link import AutoflowLink

ArgumentLinkNodeType = TypeVar("ArgumentLinkNodeType")

class ArgumentLink(AutoflowLink, Generic[ArgumentLinkNodeType]):
    model_config = ConfigDict(
        arbitrary_types_allowed=True
    )


    compressed_edges: list[tuple['ArgumentLink', ArgumentLinkNodeType]] = []
    shadow_collection: Any = None
    is_compressed_edge: bool = False

    @property
    def compressed_nodes(
        self
    ) -> list[tuple['ArgumentLink', ArgumentLinkNodeType]]:
        # Compatability
        return self.compressed_edges



import functools
import inspect
import warnings
from itertools import islice
from typing import Callable, ParamSpec, TypeVar


from . import tracer as gn
from .link import AutoflowLink

C = ParamSpec("C")
R = TypeVar("R")


def autotrace(func: Callable[C, R]) -> Callable[C, R]:
    """Function decorator that enables automatic dependency tracking

    The result is turned into :func:`traced <malevich._autoflow.tracer.traced>`
    object (if it is not already).

    Then, all arguments that are traced are linked to the result
    by calling :func:`traced.calledby`
    method with results being the 'caller'.

    The link produced is :class:`ArgumentLink <malevich.models.argument.ArgumentLink`
    with ``index`` set to the argument position and ``name`` set to the argument name.
    If the argument is passed as a keyword argument,
    and the function accepts the argument as
    **kwargs, raises a warning and does not link the argument.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # Call function to obtain resuts
        result = func(*args, **kwargs)
        result = gn.traced(result) if not isinstance(
            result, gn.traced # If the results is not traced, trace it
        ) else result
        # Obtain parameters from function signature.
        parameters = list(inspect.signature(func).parameters.values())
        # !! The function only traces positional only arguments
        # which are the ones that preceed \
        # def func(arg1, arg2, /, arg3, *, ...) <-- only arg1 and arg2 are
        # considered
        varnames = [ # filter pos-only args
            p.name for p in parameters
            if p.kind == inspect.Parameter.POSITIONAL_ONLY

        ]
        # tracing arguments, slicing up to known args
        new_edges = 0
        for i, arg in enumerate(islice(args, 0, len(varnames))):
            argument_name = varnames[i]
            if isinstance(arg, gn.traced):
                arg._autoflow.calledby(
                    result,
                    AutoflowLink(index=i, name=argument_name)
                )
                new_edges += 1
            else:
                raise ValueError(
                    f"You passed invalid argument to {func.__name__} "
                    f"at position {i}. When using processors, you may only "
                    "pass specific objects produced by Malevich operations."
                    # TODO: documentation ref
                )

        if not new_edges:
            result._autoflow._tree_ref().add_node(result)

        return result
    return wrapper


def sinktrace(func: Callable[C, R]) -> Callable[C, R]:
    """A special form of autotrace to trace functions with *args

    This decorator is applied to processors that contains
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> gn.traced:
        # Call function to obtain resuts
        result = func(*args, **kwargs)
        result = gn.traced(result) if not isinstance(result, gn.traced) else result
        parameters = list(inspect.signature(func).parameters.values())
        names = [
            p.name for p in parameters
            if p.kind in (
                inspect.Parameter.VAR_POSITIONAL,
                inspect.Parameter.POSITIONAL_ONLY
            )
        ]
        new_edges = 0
        for i, arg in enumerate(args):
            real_index = min(i, len(names) - 1)
            argument_name = names[real_index]
            if isinstance(arg, gn.traced):
                new_edges += 1
                arg._autoflow.calledby(result, AutoflowLink(
                        index=real_index,
                        name=argument_name,
                        in_sink=i >= len(names) - 1
                    )
                )
            else:
                warnings.warn(
                    "Ignoring non-traced argument in sinktrace function"
                    f" (argument index= {i})"
                )
        if not new_edges:
            result._autoflow._tree_ref().add_node(result)

        return result

    return wrapper



import ast
import json
import os
import re
import shutil
import site
import sys
from pathlib import Path
from subprocess import CalledProcessError, call
from typing import Union

import rich
import typer
from datamodel_code_generator import InputFileType, generate
from typing_extensions import Annotated

dev = typer.Typer(help="Document operations")

SECTIONS = ["heading", "input", "output", "config", "example"]


def error_exit(msg: str = "") -> None:
    rich.print(f"[bold red]{msg}[/bold red]", file=sys.stderr)
    exit(1)


TYPES = {
    "str": "string",
    "int": "integer",
    "float": "number",
    "list": "array",
    "tuple": "array",
    "dict": "object",
    "bool": "boolean",
}


def parse_type(in_type: str) -> dict:
    result = []
    types = in_type.split("|")
    for t in types:
        if "[" in t:
            search = re.search(r"(?P<PARENT>.+?)\[(?P<CHILD>.+)\]", t)
            parent = search.group("PARENT")
            if parent in TYPES.keys():
                parent = TYPES[parent]
            child = search.group("CHILD")
            result.append({"type": parent, "items": {"anyOf": parse_type(child)}})

        else:
            if t in TYPES.keys():
                t = TYPES[t]
            result.append({"type": t})
    return result


def parse_in_out(input: str = "", idx: int = 0) -> list:
    result = []
    cases = input.split("---")
    for case in cases:
        case_json = {}
        if not re.search(r"^\s+\-", case, re.MULTILINE):
            case_json["columns"] = None
            case_json["summary"] = re.search(r"([\S\s]+)", case, re.MULTILINE).group(1)
        else:
            case_json["summary"] = re.search(
                r"([\S\s]+?)^\s+\-", case, re.MULTILINE
            ).group(1)
            case_json["columns"] = []
            columns = re.findall(
                r"^\s+\-\s(?P<COLUMN_NAME>`?\w+`?+)\s+\("
                r"(?P<COLUMN_TYPE>`?\w+`?)(?P<OPTIONAL_FLAG>,\soptional)?\)"
                r"\:[\s\n\t]+(?P<COLUMN_DESCRIPTION>.+)",
                case,
                flags=re.MULTILINE,
            )
            if len(columns) == 0:
                error_exit(f"Failed to parse {SECTIONS[idx]} columns")
            for column in columns:
                case_json["columns"].append(
                    {
                        "name": column[0],
                        "type": column[1],
                        "optional": True if column[2] != "" else False,
                        "description": column[3],
                    }
                )
        result.append(case_json)
    return result


def parse_config(config: str = "") -> list:
    result = []
    columns = re.findall(
        r"^\s+- (?P<FIELD_NAME>`?\w+`?)\:\s(?P<FIELD_TYPE>`?[|\w\[\{\<\]\}\>]+`?)"
        r"(?P<DEFAULT_CLAUSE>\,\sdefault (?P<DEFAULT_VALUE>.+))?\.[\s\n\t]+"
        r"(?P<FIELD_DESCRIPTION>.+)\.$",
        config,
        re.MULTILINE,
    )
    if len(columns) != len(re.findall(r"^\s+- ", config, re.MULTILINE)):
        in_doc = len(re.findall(r"^\s+- ", config, re.MULTILINE))
        error_exit(
            "Failed to parse config columns\n"
            f"Number of columns in doc (started with ' - ') is {in_doc}), "
            f"Number of columns parsed is {len(columns)}"
        )
    for column in columns:
        result.append(
            {
                "name": column[0],
                "type": column[1],
                "default": None if column[3] == "" else column[3],
                "description": column[4],
            }
        )
    return result


@dev.command("list-procs", help="List all processors and paths to their files")
def list_procs(
    path=typer.Argument(
        ".", help="Directory to extract processors from", show_default=False
    ),
    out: Annotated[
        Union[str, None],
        typer.Option("--out", help="File to save results in", show_default=False),
    ] = None,
    verbose: Annotated[
        bool,
        typer.Option(
            "--verbose", "-v", help="If enabled, will print result JSON to stdout"
        ),
    ] = False,
) -> str:
    data = []
    if os.path.isdir(path):
        for root, _, files in os.walk(path):
            for file in files:
                # Parsing python files
                if file.endswith(".py"):
                    with open(os.path.join(root, file)) as f:
                        # Parsing AST
                        tree = ast.parse(f.read())
                        for node in ast.walk(tree):
                            # Looking for functions with @processor decorator
                            if isinstance(node, ast.FunctionDef) or isinstance(
                                node, ast.AsyncFunctionDef
                            ):
                                function_name = node.name
                                decorators = [
                                    # id may be absent?
                                    d.func.id
                                    for d in node.decorator_list
                                    if isinstance(d, ast.Call)
                                ]
                                if "processor" in decorators:
                                    data.append(
                                        {
                                            "name": function_name,
                                            "path": os.path.abspath(f"{root}/{file}"),
                                        }
                                    )
    else:
        with open(path) as f:
            tree = ast.parse(f.read())
            for node in ast.walk(tree):
                if isinstance(node, ast.FunctionDef) or isinstance(
                    node, ast.AsyncFunctionDef
                ):
                    function_name = node.name
                    decorators = [
                        d.func.id
                        for d in node.decorator_list
                        if isinstance(d, ast.Call)
                    ]
                    if "processor" in decorators:
                        data.append(
                            {
                                "name": function_name,
                                "path": os.path.abspath(path),
                            }
                        )

    if out is not None:
        with open(out, "w") as f:
            f.write(json.dumps(data))
    if verbose:
        rich.print(json.dumps(data))
    return json.dumps(data)


@dev.command("get-doc", help="Get process docstring")
def get_processor_docstring(
    name=typer.Argument(
        ...,
        help="Processor name",
    ),
    path=typer.Argument(
        ...,
        help="Path to the file",
    ),
    out: Annotated[
        Union[str, None],
        typer.Option("--out", help="File to save results in", show_default=False),
    ] = None,
    verbose: Annotated[
        bool,
        typer.Option(
            "--verbose", "-v", help="If enabled, will print result docstring to stdout"
        ),
    ] = False,
) -> str:
    if not os.path.exists(path):
        error_exit(f"Can not open {path}. No such file exists")

    with open(path) as f:
        tree = ast.parse(f.read())
        for node in ast.walk(tree):
            if (
                isinstance(node, ast.FunctionDef)
                or isinstance(node, ast.AsyncFunctionDef)
            ) and node.name == name:
                doc = ast.get_docstring(node)
                if out is not None:
                    open(out, "w").write(doc)
                if verbose:
                    rich.print(doc)
                return doc

    rich.print(
        "Could not find processor inside a file. "
        "Make sure you have provided correct arguments"
    )
    exit(1)


@dev.command("parse-doc", help="Parse processor docstring")
def parse_docstring(
    doc=typer.Argument(..., help="docstring"),
    file: Annotated[
        bool,
        typer.Option("--file", "-f", help="If enabled, will read from argument string"),
    ] = False,
    out: Annotated[
        Union[str, None],
        typer.Option("--out", help="File to save results in", show_default=False),
    ] = None,
    verbose: Annotated[
        bool,
        typer.Option(
            "--verbose", "-v", help="If enabled, will print result JSON to stdout"
        ),
    ] = False,
) -> str:
    idx = 0
    if file:
        doc = open(doc).read()

    if not doc:
        error_exit("Missing documentation")

    doc = doc.replace("\n", "\\n ")
    try:
        user = re.search(r"(?P<USER>[\S\s]+)-----(?P<DEV>[\S\s]*)", doc).group("USER")
    except Exception:
        error_exit("Failed to divide doc into User/Dev sections")

    result = {}

    try:
        heading = (
            re.search(r"(?P<HEADING>[\S\s]+?)(##|$)", user)
            .group("HEADING")
            .replace("\\n ", "\n")
        )
        result["heading"] = heading
        idx += 1

        input = (
            re.search(r"## Input:\\n(?P<INPUT>[\S\s]+?)(##|$)", user)
            .group("INPUT")
            .replace("\\n ", "\n")
        )
        result["input_section"] = parse_in_out(input, idx)
        idx += 1

        output = (
            re.search(r"## Output:\\n(?P<OUTPUT>[\S\s]+?)(#|$)", user)
            .group("OUTPUT")
            .replace("\\n ", "\n")
        )
        result["output_section"] = parse_in_out(output, idx)

        if "## Configuration" in doc:
            idx += 1
            config = (
                re.search(r"## Configuration:\\n(?P<CONFIG>[\S\s]+?)(#|$)", user)
                .group("CONFIG")
                .replace("\\n ", "\n")
            )
            result["configuration"] = parse_config(config)
        if "## Example" in doc:
            idx += 1
            example = (
                re.search(r"## Example:\\n(?P<EXAMPLE>[\S\s]+?)(#|$)", user)
                .group("EXAMPLE")
                .replace("\\n ", "\n")
            )
            result["example"] = example
    except Exception:
        error_exit(f"Docstring doesn't suit the rules in {SECTIONS[idx]} section")

    if out:
        open(out, "w").write(json.dumps(result))
    if verbose:
        rich.print(json.dumps(result))
    return json.dumps(result)


@dev.command("install-lib-hook", help="Configure git hooks path")
def intstall_hook(
    path=typer.Option(
        ".githooks/",
        help="Folder with hooks",
        show_default=False,
    ),
) -> None:
    if not os.path.exists(path):
        error_exit(
            f"Cannot find {path}. "
            "Make sure you've provided a valid path, and run from the right directory",
        )

    try:
        call(["git", "config", "core.hooksPath", path])
    except CalledProcessError:
        error_exit("Failed to execute the process")


@dev.command("make-configs", help="Create Context configurations for processors")
def make_config(
    path=typer.Argument(..., help="Path to start directory"),
    verbose: Annotated[
        bool,
        typer.Option(
            "--verbose", "-v", help="If enabled, will print result JSON to stdout"
        ),
    ] = False,
) -> None:
    data = json.loads(list_procs(path))
    by_path = {}
    for d in data:
        names: list = by_path.get(d["path"], [])
        names.append(d["name"])
        by_path[d["path"]] = names

    for path_ in by_path.keys():
        modules = []
        models_path = None
        rich.print(f"Processing {path_}...")
        for name_ in by_path[path_]:
            doc = get_processor_docstring(name_, path_)
            schema = json.loads(parse_docstring(doc))
            if "configuration" in schema.keys() and len(schema["configuration"]) > 0:
                schema_path = os.path.join(
                    os.path.dirname(path_), "models", f"{name_}_model.json"
                )
                schema_model = os.path.join(
                    os.path.dirname(path_), "models", f"{name_}_model.py"
                )
                models_path = os.path.dirname(schema_model)
                if not os.path.exists(os.path.dirname(schema_path)):
                    os.mkdir(os.path.dirname(schema_path))
                props = {}
                required = []
                for conf in schema["configuration"]:
                    conf_json = {}
                    conf_name = conf["name"].strip("`\"'")
                    conf_json["anyOf"] = parse_type(conf["type"])
                    if conf["default"] is not None:
                        conf_json["default"] = eval(conf["default"])
                    else:
                        required.append(conf_name)
                    if "description" in conf.keys():
                        conf_json["description"] = conf["description"]
                    props[conf_name] = conf_json

                schema = json.dumps(
                    {"type": "object", "properties": props, "required": required}
                )

                with open(schema_path, "w") as f:
                    f.write(schema)

                generate(
                    Path(schema_path),
                    output=Path(schema_model),
                    class_name=name_,
                    input_file_type=InputFileType.JsonSchema,
                    disable_timestamp=True
                )
                model = open(schema_model).read()
                model = re.sub(
                    r"^class", "@scheme()\nclass", model, flags=re.MULTILINE
                ).replace(
                    "from __future__ import annotations",
                    "from __future__ import annotations\n"
                    "from malevich.square import scheme",
                )

                with open(schema_model, "w") as f:
                    f.write(model)
                os.remove(schema_path)

                modules.append(
                    {
                        "name": name_,
                        "module": f"{name_}_model",
                        "classname": re.search(
                            r"^class (?P<CLASS>\w+)\(", model, flags=re.MULTILINE
                        ).group("CLASS"),
                    }
                )
        if models_path is not None:
            init_file = os.path.join(models_path, "__init__.py")
            Path(init_file).touch()
            file = open(init_file).read()
            procs = open(path_).read()
            for module in modules:
                search_import = re.search(
                    rf"^from .{module['module']} import {module['classname']}",
                    file,
                    flags=re.MULTILINE,
                )
                if search_import is None:
                    if verbose:
                        rich.print(f"Added {module['classname']} to __init__.py")
                    file += f"from .{module['module']} import {module['classname']}\n"
                if (
                    re.search(
                        rf"^from .models import {module['classname']}$",
                        procs,
                        flags=re.MULTILINE,
                    )
                    is None
                ):
                    if verbose:
                        rich.print(f"Added{module['classname']} to {path_}")
                    procs = f"from .models import {module['classname']}\n" + procs

                search_proc = re.search(
                    rf"(?P<DEF>^(async )?def {module['name']}"
                    r"\([\s\S]*? Context)(\[\w+\])?",
                    procs,
                    flags=re.MULTILINE,
                )
                group = search_proc.group("DEF").replace(
                    "Context", f"Context[{module['classname']}]"
                )
                procs = re.sub(
                    rf"(?P<DEF>^(async )?def {module['name']}"
                    r"\([\s\S]*? Context(\[\w+\])?)",
                    group,
                    procs,
                    flags=re.MULTILINE,
                )
            with open(init_file, "w") as f:
                f.write(file)
            with open(path_, "w") as f:
                f.write(procs)


@dev.command("procs-info", help="Get processors info in JSON format")
def procs_info(
    path=typer.Argument("./", show_default=False),
    out: Annotated[
        Union[str, None],
        typer.Option("--out", help="File to save results in", show_default=False),
    ] = None,
    verbose: Annotated[
        bool,
        typer.Option(
            ...,
            "--verbose",
            "-v",
            help="Verbose mode",
        ),
    ] = False,
) -> str | None:
    procs = json.loads(list_procs(path))
    if len(procs) == 0:
        rich.print("No procs found.")
        return
    info = []
    for p in procs:
        if verbose:
            rich.print(
                f"Processing [bold]{p['name']}[/bold] "
                f"from [italic]{p['path']}[/italic]"
            )
        doc = get_processor_docstring(p["name"], p["path"])
        info.append(json.loads(parse_docstring(doc)))

    if verbose:
        rich.print(info)
    if out is not None:
        with open(out, "w") as f:
            f.write(json.dumps(info))
    return json.dumps(info)


@dev.command("in-app-install", help="Install malevich inside the app")
def in_app_install() -> None:
    if os.getcwd() == "/julius":
        pkg_ = site.getsitepackages()[0]
        shutil.rmtree(os.path.join(pkg_, "malevich", "square"))
        shutil.copytree(
            "/julius/malevich/square", os.path.join(pkg_, "malevich", "square")
        )
        shutil.rmtree("/julius/malevich")
        shutil.move(os.path.join(pkg_, "malevich"), "/julius/malevich")
    else:
        error_exit("You are running command outside of the app")

@dev.command("get-regex", help="Get regexps by which we check documentation.")
def get_regexp() -> None:
    rich.print(
        "Input/Output columns:",
        "^\\s+\\-\\s(?P<COLUMN_NAME>`?\\w+`?+)\\s+\\("
        "(?P<COLUMN_TYPE>`?\\w+`?)(?P<OPTIONAL_FLAG>,\\soptional)?\\)"
        "\\:[\\s\\n\\t]+(?P<COLUMN_DESCRIPTION>.+)",
        "\n\n"
    )
    rich.print(
        "Config columns:",
        "^\\s+- (?P<FIELD_NAME>`?\\w+`?)\\:\\s"
        "(?P<FIELD_TYPE>`?[|\\w\\[\\{\\<\\]\\}\\>]+`?)"
        "(?P<DEFAULT_CLAUSE>\\,\\sdefault (?P<DEFAULT_VALUE>.+))?\\.[\\s\\n\\t]+"
        "(?P<FIELD_DESCRIPTION>.+)\\.$"
    )



from .env import EnvManager
from .suite import (
    FlowTestEnv,
    FlowTestSuite,
    ImageDependency,
    SpaceDependency,
    clean_env
)
from malevich.models.installers.image import ImageDependency, ImageOptions
from malevich.models.installers.space import SpaceDependency, SpaceDependencyOptions



import argparse
from datetime import datetime
import os
import sys
import dill

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..')))

parser = argparse.ArgumentParser()
parser.add_argument('session_id', type=int)
args = parser.parse_args()

os.environ['_MALEVICH_NO_FINALIZE'] = 'true'

from malevich._db import Read, Write
from malevich._db.schema.artifacts import Artifact, MasterSession, Session
from malevich.path import Paths

with Write() as db_ref:
    session = db_ref.get(Session, args.session_id)
    # session_artifacts = db_ref.query(Artifact).filter(
    #     Artifact.session_id == args.session_id
    # ).all()

    # declared_flows = set()
    # files = set()

    # for artifact in session_artifacts:
    #     artifact.payload = dill.loads(artifact.payload)
    #     if artifact.payload['type'] == 'flow_declaration':
    #         declared_flows.add(artifact.payload['reverse_id'])
    #     if artifact.payload['type'] == 'file_content':
    #         files.add(artifact.payload['file'])

    # recent_artifacts = db_ref.query(Artifact).all()
    session.end = datetime.now()
    db_ref.commit()



from typing import Generic, TypeVar

from .base import BaseRef

CreateFn = TypeVar('CreateFn', bound=callable)
DeleteFn = TypeVar('DeleteFn', bound=callable)
UpdateFn = TypeVar('UpdateFn', bound=callable)
GetFn = TypeVar('GetFn', bound=callable)
ListFn = TypeVar('ListFn', bound=callable)
GetLinkFn = TypeVar('GetLinkFn', bound=callable)
PostLinkFn = TypeVar('PostLinkFn', bound=callable)

class AssetRef(
    Generic[CreateFn, DeleteFn, UpdateFn, GetFn, ListFn, GetLinkFn, PostLinkFn],
    BaseRef[CreateFn, DeleteFn, UpdateFn, GetFn, ListFn],
):
    def __init__(
        self,
        name: str,
        create: CreateFn | None = None,
        delete: DeleteFn | None = None,
        update: UpdateFn | None = None,
        get: GetFn | None = None,
        list: ListFn | None = None,
        get_link: GetLinkFn | None = None,
        post_link: PostLinkFn | None = None,
    ) -> None:
        super().__init__(name, create, delete, update, get, list)
        self.get_link = get_link
        self.post_link = post_link



import os
import tempfile
import zipfile

import requests


def upload_zip_asset(
    url: str,
    file: os.PathLike | None = None,
    files: list[os.PathLike] | None = None,
    folder: os.PathLike | None = None,
):
    """Uploads a zip asset to the platform

    Args:
        file (os.PathLike, optional): Path to the file to upload. Defaults to None.
        files (list[os.PathLike], optional): List of paths to the files to upload. Defaults to None.
        folder (os.PathLike, optional): Path to the folder to upload. Defaults to None.
    """
    if file is not None:
        response = requests.post(url, data=open(file, 'rb').read())
    elif folder is not None:
        files = [os.path.join(folder, f) for f in os.listdir(folder)]
        with tempfile.TemporaryFile(mode='+rb') as temp_file:
            with zipfile.ZipFile(temp_file, 'w') as zip_file:
                for f in files:
                    zip_file.write(f)
            temp_file.seek(0)
            response: requests.Response = requests.post(url, data=temp_file.read(), params={
                'zip': True
            })

    response.raise_for_status()
    return response.raw
